# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import functools
import inspect
from dataclasses import asdict, dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    NoReturn,
    Optional,
    Protocol,
    TypeVar,
    Union,
    cast,
    overload,
    runtime_checkable,
)
from uuid import UUID

import structlog
from nwastdlib import const, identity
from structlog.stdlib import BoundLogger
from structlog.threadlocal import tmp_bind

from orchestrator.config.assignee import Assignee
from orchestrator.db import EngineSettingsTable, db, transactional
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import (
    ErrorDict,
    FormGenerator,
    InputFormGenerator,
    InputStepFunc,
    State,
    StateInputFormGenerator,
    StateInputStepFunc,
    StateSimpleInputFormGenerator,
    StepFunc,
    strEnum,
)
from orchestrator.utils.docs import make_workflow_doc
from orchestrator.utils.errors import error_state_to_dict
from orchestrator.utils.state import form_inject_args, inject_args

logger = structlog.get_logger(__name__)


StepLogFunc = Callable[["ProcessStat", "Step", "Process"], "Process"]
StepLogFuncInternal = Callable[["Step", "Process"], "Process"]
StepToProcessFunc = Callable[[State], "Process"]


@runtime_checkable
class Step(Protocol):
    __name__: str
    __qualname__: str
    name: str
    form: Optional[InputFormGenerator]
    assignee: Optional[Assignee]

    def __call__(self, state: State) -> Process:
        ...


@runtime_checkable
class Workflow(Protocol):
    __name__: str
    __qualname__: str
    name: str
    description: str
    initial_input_form: Optional[InputFormGenerator] = None
    target: Optional[Target] = None
    steps: StepList

    def __call__(self) -> NoReturn:
        ...


def make_step_function(
    f: Callable, name: str, form: Optional[InputFormGenerator] = None, assignee: Optional[Assignee] = Assignee.SYSTEM
) -> Step:
    step_func = cast(Step, f)

    step_func.name = name
    step_func.form = form
    step_func.assignee = assignee
    return step_func


class StepList(List[Step]):
    """Wraps around a primitive list of `Step` to provide a "list" with associative `append` (or its alias: `>>`).

    >>> one = step("one")(dict)
    >>> two = step("two")(dict)
    >>> three = step("three")(dict)

    >>> empty = StepList([])
    >>> empty >> empty
    StepList []

    >>> empty == empty >> empty
    True

    >>> str(empty >> one)
    'StepList [one]'

    >>> (begin >> one >> two) >> three == begin >> one >> (begin >> two >> three)
    True
    """

    def map(self, f: Callable) -> StepList:
        return StepList(map(f, self))

    @overload
    def __getitem__(self, i: int) -> Step:
        ...

    @overload  # noqa: F811
    def __getitem__(self, i: slice) -> StepList:
        ...

    def __getitem__(self, i: Union[int, slice]) -> Union[Step, StepList]:  # noqa: F811
        retval: Union[Step, List[Step]] = super().__getitem__(i)
        if isinstance(retval, list):
            # ensure we return a StepList and not a regular list.
            retval = type(self)(retval)
        return retval

    def __rshift__(self, other: Union[StepList, Step]) -> StepList:
        if isinstance(other, Step):
            return StepList([*self, other])
        elif isinstance(other, StepList):
            return StepList([*self, *other])
        elif hasattr(other, "__name__"):  # type:ignore
            raise ValueError(
                f"Expected @step decorated function or type Step or StepList, got {type(other)} with name {other.__name__} instead."
            )
        else:
            raise ValueError(f"Expected @step decorated function or type Step or StepList, got {type(other)} instead.")

    def __str__(self) -> str:
        return f"StepList [{', '.join(x.name for x in self)}]"

    def __repr__(self) -> str:
        return f"StepList [{', '.join(repr(x) for x in self)}]"


def _handle_simple_input_form_generator(f: StateInputStepFunc) -> StateInputFormGenerator:
    if inspect.isgeneratorfunction(f):
        return cast(StateInputFormGenerator, f)
    if inspect.isgenerator(f):
        raise ValueError("Got a generator object instead of function, this is not correct")
    else:
        # If f is a SimpleInputFormGenerator convert to new style generator function
        def form_generator(state: State) -> FormGenerator:
            user_input = yield cast(StateSimpleInputFormGenerator, f)(state)
            return user_input.dict()

        return form_generator


def make_workflow(
    f: Callable,
    description: str,
    initial_input_form: Optional[InputStepFunc],
    target: Optional[Target],
    steps: StepList,
) -> Workflow:
    @functools.wraps(f)
    def wrapping_function() -> NoReturn:
        raise Exception("This function should not be executed")

    wrapping_function = cast(Workflow, wrapping_function)

    wrapping_function.name = f.__name__  # default, will be changed by LazyWorkflowInstance
    wrapping_function.description = description

    if initial_input_form is None:
        # We always need a form to prevent starting a workflow when no input is needed.
        # This would happen on first post that is used to retrieve the first form page
        initial_input_form = cast(InputStepFunc, const(FormPage))

    wrapping_function.initial_input_form = _handle_simple_input_form_generator(initial_input_form)
    wrapping_function.target = target
    wrapping_function.steps = steps

    wrapping_function.__doc__ = make_workflow_doc(wrapping_function)

    return wrapping_function


def step(name: str) -> Callable[[StepFunc], Step]:
    """Mark a function as a workflow step."""

    def decorator(func: StepFunc) -> Step:
        @functools.wraps(func)
        def wrapper(state: State) -> Process:
            with tmp_bind(logger, func=func.__qualname__) as log:

                step_in_inject_args = inject_args(func)

                try:
                    with transactional(db, log):
                        result = step_in_inject_args(state)
                        return Success(result)
                except Exception as ex:
                    log.warning("Step failed", exc_info=ex)
                    return Failed(ex)

        return make_step_function(wrapper, name)

    return decorator


def retrystep(name: str) -> Callable[[StepFunc], Step]:
    """Mark a function as a retryable workflow step.

    If this step fails it goes to `Waiting` were it will be retried periodically. If it `Success` it acts as a normal
    step.
    """

    def decorator(func: StepFunc) -> Step:
        @functools.wraps(func)
        def wrapper(state: State) -> Process:
            with tmp_bind(logger, func=func.__qualname__) as log:
                step_in_inject_args = inject_args(func)

                try:
                    with transactional(db, log):
                        result = step_in_inject_args(state)
                        return Success(result)
                except Exception as ex:
                    return Waiting(ex)

        return make_step_function(wrapper, name)

    return decorator


def inputstep(name: str, assignee: Optional[Assignee] = None) -> Callable[[InputStepFunc], Step]:
    """Add user input step to workflow.

    IMPORTANT: In contrast to other workflow steps, the `@inputstep` wrapped function will not run in the
    workflow engine! This means that it should never do any changes in the database and external systems!

    Example::

        @inputstep("User step", assignee=Assignee.NOC)
        def user_step(state: State) -> FormGenerator:
            class Form(FormPage):
                name: str
            user_input = yield Form
            return {**user_input.dict(), "some extra key": True}

    """

    def decorator(func: InputStepFunc) -> Step:
        def wrapper(state: State) -> FormGenerator:
            form_generator_in_form_inject_args = form_inject_args(func)

            form_generator = _handle_simple_input_form_generator(form_generator_in_form_inject_args)

            return form_generator(state)

        @functools.wraps(func)
        def suspend(state: State) -> Process:
            return Suspend(state)

        return make_step_function(suspend, name, wrapper, assignee)

    return decorator


def _purestep(name: str) -> Callable[[StepToProcessFunc], StepList]:
    """Part of workflow "DSL" to map a `state -> Process state` function into a workflow step."""

    def _purestep(f: StepToProcessFunc) -> StepList:
        return StepList([make_step_function(f, name)])

    return _purestep


def conditional(p: Callable[[State], bool]) -> Callable[..., StepList]:
    """Use a predicate to control whether or not a step is run."""

    def _conditional(steps_or_func: Union[StepList, Step]) -> StepList:
        if isinstance(steps_or_func, Step):
            steps = StepList([steps_or_func])
        else:
            steps = steps_or_func

        def wrap(step: Step) -> Step:
            @functools.wraps(step)
            def wrapper(state: State) -> Process:
                return step(state) if p(state) else Skipped(state)

            return make_step_function(wrapper, step.name, step.form, step.assignee)

        return steps.map(wrap)

    return _conditional


def steplens(get: Callable[[State], State], set: Callable[[State], Callable[[State], State]]) -> Callable[[Step], Step]:
    """Update a step list to zoom its input state using get and update its output state using set."""

    def wrap(step: Step) -> Step:
        @functools.wraps(step)
        def wrapper(state: State) -> Process:
            sub_state = get(state)

            result: Process = step(sub_state)

            if result.isfailed() or result.iswaiting():
                return result
            else:
                return result.map(set(state))

        return make_step_function(wrapper, step.name, step.form, step.assignee)

    return wrap


def focussteps(key: str) -> Callable[[Union[Step, StepList]], StepList]:
    """Return a function that maps `steplens` over `steps`, getting and setting a single key."""

    def zoom(steps_or_func: Union[Step, StepList]) -> StepList:
        if isinstance(steps_or_func, Step):
            steps = StepList([steps_or_func])
        else:
            steps = steps_or_func

        def get(state: State) -> State:
            return state.get(key, {})

        def set(state: State) -> Callable[[State], State]:
            return lambda substate: {**state, key: substate}

        return steps.map(steplens(get, set))

    return zoom


def workflow(
    description: str, initial_input_form: Optional[InputStepFunc] = None, target: Optional[Target] = None
) -> Callable[[Callable[[], StepList]], Workflow]:
    """Transform an initial_input_form and a step list into a workflow.

    Use this for other workflows. For create workflows use :func:`create_workflow`

    Example::

        @workflow("create service port")
        def create_service_port():
            init
            << do_something
            << done
    """
    if initial_input_form is None:
        initial_input_form_in_form_inject_args = None
    else:
        initial_input_form_in_form_inject_args = form_inject_args(initial_input_form)

    def _workflow(f: Callable[[], StepList]) -> Workflow:
        return make_workflow(f, description, initial_input_form_in_form_inject_args, target, f())

    return _workflow


@dataclass
class ProcessStat:
    pid: UUID
    workflow: Workflow
    state: Process
    log: StepList
    current_user: str

    def update(self, **vs: Any) -> ProcessStat:
        """Update ProcessStat.

        >>> pstat = ProcessStat('', None, {}, [], "")
        >>> pstat.update(state={"a": "b"})
        ProcessStat(pid='', workflow=None, state={'a': 'b'}, log=[], current_user='')
        """
        return ProcessStat(**{**asdict(self), **vs})


S = TypeVar("S")
F = TypeVar("F")


class ProcessStatus(strEnum):
    CREATED = "created"
    RUNNING = "running"
    SUSPENDED = "suspended"
    WAITING = "waiting"
    ABORTED = "aborted"
    FAILED = "failed"
    API_UNAVAILABLE = "api_unavailable"
    INCONSISTENT_DATA = "inconsistent_data"
    COMPLETED = "completed"


class StepStatus(strEnum):
    SUCCESS = "success"
    SKIPPED = "skipped"
    SUSPEND = "suspend"
    WAITING = "waiting"
    FAILED = "failed"
    ABORT = "abort"
    COMPLETE = "complete"


class Process(Generic[S]):
    def __init__(self, s: S):
        self.s = s

    def map(self, f: Callable[[S], S]) -> Process[S]:
        """Apply a function to the process.

        >>> inc = lambda n: n + 1

        >>> Success(1).map(inc)
        Success 2

        >>> Skipped(1).map(inc)
        Skipped 2

        >>> Suspend(1).map(inc)
        Suspend 2

        >>> Waiting(1).map(inc)
        Waiting 2

        >>> Abort(1).map(inc)
        Abort 2

        >>> Failed(1).map(inc)
        Failed 2

        >>> Complete(1).map(inc)
        Complete 2
        """

        def g(x: S) -> Process[S]:
            Self = self.__class__
            return Self(f(x))

        return self._fold(g, g, g, g, g, g, g)

    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        """Unwrap the state from the Process category.

        >>> Success('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Success 'a'

        >>> Skipped('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Skipped 'a'

        >>> Suspend('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Suspend 'a'

        >>> Waiting('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Waiting 'a'

        >>> Abort('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Abort 'a'

        >>> Failed('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Failed 'a'

        >>> Complete('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Complete 'a'

        >>> Process('a')._fold(Success, Skipped, Suspend, Waiting, Abort, Failed, Complete)
        Traceback (most recent call last):
            ...
        NotImplementedError: Abstract function `_fold` must be implemented by the type constructor
        """
        raise NotImplementedError("Abstract function `_fold` must be implemented by the type constructor")

    def unwrap(self) -> S:
        """Get unwrapped state.

        >>> Success('a').unwrap()
        'a'

        >>> Skipped('a').unwrap()
        'a'

        >>> Suspend('a').unwrap()
        'a'

        >>> Waiting('a').unwrap()
        'a'

        >>> Abort('a').unwrap()
        'a'

        >>> Failed('a').unwrap()
        'a'

        >>> Complete('a').unwrap()
        'a'
        """
        return self._fold(identity, identity, identity, identity, identity, identity, identity)  # type: ignore

    def issuccess(self) -> bool:
        """Test if this instance is Success.

        >>> Success('a').issuccess()
        True

        >>> Skipped('a').issuccess()
        False

        >>> Suspend('a').issuccess()
        False

        >>> Waiting('a').issuccess()
        False

        >>> Abort('a').issuccess()
        False

        >>> Failed('a').issuccess()
        False

        >>> Complete('a').issuccess()
        False
        """
        return self._fold(
            const(True), const(False), const(False), const(False), const(False), const(False), const(False)
        )

    def isskipped(self) -> bool:
        """Test if this instance is Skipped.

        >>> Success('a').isskipped()
        False

        >>> Skipped('a').isskipped()
        True

        >>> Suspend('a').isskipped()
        False

        >>> Waiting('a').isskipped()
        False

        >>> Abort('a').isskipped()
        False

        >>> Failed('a').isskipped()
        False

        >>> Complete('a').isskipped()
        False
        """
        return self._fold(
            const(False), const(True), const(False), const(False), const(False), const(False), const(False)
        )

    def issuspend(self) -> bool:
        """Test if this instance is Suspend.

        >>> Success('a').issuspend()
        False

        >>> Skipped('a').issuspend()
        False

        >>> Suspend('a').issuspend()
        True

        >>> Waiting('a').issuspend()
        False

        >>> Abort('a').issuspend()
        False

        >>> Failed('a').issuspend()
        False

        >>> Complete('a').issuspend()
        False
        """
        return self._fold(
            const(False), const(False), const(True), const(False), const(False), const(False), const(False)
        )

    def isabort(self) -> bool:
        """Test if this instance is Abort.

        >>> Success('a').isabort()
        False

        >>> Skipped('a').isabort()
        False

        >>> Suspend('a').isabort()
        False

        >>> Waiting('a').isabort()
        False

        >>> Abort('a').isabort()
        True

        >>> Failed('a').isabort()
        False

        >>> Complete('a').isabort()
        False
        """
        return self._fold(
            const(False), const(False), const(False), const(False), const(True), const(False), const(False)
        )

    def iswaiting(self) -> bool:
        """Test if this instance is Waiting.

        >>> Success('a').iswaiting()
        False

        >>> Skipped('a').iswaiting()
        False

        >>> Suspend('a').iswaiting()
        False

        >>> Waiting('a').iswaiting()
        True

        >>> Abort('a').iswaiting()
        False

        >>> Failed('a').iswaiting()
        False

        >>> Complete('a').iswaiting()
        False
        """
        return self._fold(
            const(False), const(False), const(False), const(True), const(False), const(False), const(False)
        )

    def isfailed(self) -> bool:
        """Test if this instance is Waiting.

        >>> Success('a').isfailed()
        False

        >>> Skipped('a').isfailed()
        False

        >>> Suspend('a').isfailed()
        False

        >>> Waiting('a').isfailed()
        False

        >>> Abort('a').isfailed()
        False

        >>> Failed('a').isfailed()
        True

        >>> Complete('a').isfailed()
        False
        """
        return self._fold(
            const(False), const(False), const(False), const(False), const(False), const(True), const(False)
        )

    def iscomplete(self) -> bool:
        """Test if this instance is Complete.

        >>> Success('a').iscomplete()
        False

        >>> Skipped('a').iscomplete()
        False

        >>> Suspend('a').iscomplete()
        False

        >>> Waiting('a').iscomplete()
        False

        >>> Abort('a').iscomplete()
        False

        >>> Failed('a').iscomplete()
        False

        >>> Complete('a').iscomplete()
        True
        """
        return self._fold(
            const(False), const(False), const(False), const(False), const(False), const(False), const(True)
        )

    def __eq__(self, other: object) -> bool:
        """Test two instances for equality.

        >>> Success('a') == Success('a')
        True

        >>> Success('a') != Success('b')
        True

        >>> Success('a') != Suspend('a')
        True

        >>> Success('a') != Waiting('a')
        True

        >>> Suspend('a') != Abort('a')
        True

        >>> Success('a') != 'a'
        True
        """
        return self.__class__ == other.__class__ and self.s == cast(Process, other).s

    @property
    def status(self) -> StepStatus:
        """Show status.

        >>> Success({}).status
        <StepStatus.SUCCESS: 'success'>

        >>> Skipped({}).status
        <StepStatus.SKIPPED: 'skipped'>

        >>> Suspend({}).status
        <StepStatus.SUSPEND: 'suspend'>

        >>> Waiting({}).status
        <StepStatus.WAITING: 'waiting'>

        >>> Abort({}).status
        <StepStatus.ABORT: 'abort'>

        >>> Failed({}).status
        <StepStatus.FAILED: 'failed'>

        >>> Complete({}).status
        <StepStatus.COMPLETE: 'complete'>
        """
        return StepStatus[self.__class__.__name__.upper()]

    @staticmethod
    def from_status(status: StepStatus, state: S) -> Optional[Process]:
        """Make Process based on status and state.

        >>> Process.from_status('success', {})
        Success {}

        >>> Process.from_status('skipped', {})
        Skipped {}

        >>> Process.from_status('suspend', {})
        Suspend {}

        >>> Process.from_status('waiting', {})
        Waiting {}

        >>> Process.from_status('abort', {})
        Abort {}

        >>> Process.from_status('failed', {})
        Failed {}

        >>> Process.from_status('complete', {})
        Complete {}

        >>> Process.from_status('unknown', {})

        """
        status_class = _STATUSES.get(status)
        return status_class(state) if status_class else None

    @property
    def overall_status(self) -> ProcessStatus:
        """Show overall status of process or task.

        >>> Success({}).overall_status
        <ProcessStatus.RUNNING: 'running'>

        >>> Skipped({}).overall_status
        <ProcessStatus.RUNNING: 'running'>

        >>> Suspend({}).overall_status
        <ProcessStatus.SUSPENDED: 'suspended'>

        >>> Waiting({}).overall_status
        <ProcessStatus.WAITING: 'waiting'>

        >>> Abort({}).overall_status
        <ProcessStatus.ABORTED: 'aborted'>

        >>> Failed({}).overall_status
        <ProcessStatus.FAILED: 'failed'>

        >>> Complete({}).overall_status
        <ProcessStatus.COMPLETED: 'completed'>
        """
        return self._fold(
            const(ProcessStatus.RUNNING),
            const(ProcessStatus.RUNNING),
            const(ProcessStatus.SUSPENDED),
            const(ProcessStatus.WAITING),
            const(ProcessStatus.ABORTED),
            const(ProcessStatus.FAILED),
            const(ProcessStatus.COMPLETED),
        )

    def __repr__(self) -> str:
        """Show self.

        >>> repr(Success({}))
        'Success {}'

        >>> repr(Skipped({}))
        'Skipped {}'

        >>> repr(Suspend({}))
        'Suspend {}'

        >>> repr(Waiting({}))
        'Waiting {}'

        >>> repr(Abort({}))
        'Abort {}'

        >>> repr(Failed({}))
        'Failed {}'

        >>> repr(Complete({}))
        'Complete {}'
        """
        name = self.__class__.__name__
        return f"{name} {self.s!r}"

    def on_success(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Success.

        >>> def assign_b(d):
        ...     d["b"] = 2
        ...     return d
        >>> Success({"a":1}).on_success(assign_b)
        Success {'a': 1, 'b': 2}

        >>> Failed({"a": 1}).on_success(assign_b)
        Failed {'a': 1}
        """
        return self.map(f) if self.issuccess() else self

    def on_skipped(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Skipped."""
        return self.map(f) if self.isskipped() else self

    def on_suspend(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Suspend."""
        return self.map(f) if self.issuspend() else self

    def on_waiting(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Waiting."""
        return self.map(f) if self.iswaiting() else self

    def on_abort(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Abort."""
        return self.map(f) if self.isabort() else self

    def on_failed(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Failed."""
        return self.map(f) if self.isfailed() else self

    def on_complete(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Complete."""
        return self.map(f) if self.iscomplete() else self

    def execute_step(self, step: Callable[[S], Process[S]]) -> Process[S]:
        """Execute a step transition based on the a step function.

        A step can only be executed if the current state is success or skipped.

        >>> Success({"a":1}).execute_step(lambda s: Success(s))
        Success {'a': 1}
        >>> Success({"a":1}).execute_step(lambda s: Failed(s))
        Failed {'a': 1}

        >>> Waiting({"a":1}).execute_step(lambda s: Failed(s))
        Waiting {'a': 1}

        """

        return self._fold(step, step, Suspend, Waiting, Abort, Failed, Complete)

    def abort(self) -> Process[S]:
        """Abort process.

        Always works except for completed processes
        """
        return self._fold(Abort, Abort, Abort, Abort, Abort, Abort, Complete)

    def resume(self, resume_suspend: Callable[[Process[S]], Process[S]]) -> Process[S]:
        """Resume process.

        Cannot resume Abort or Complete states

        Args:
            resume_suspend: function to call on resuming a suspended state. Might fail and determine the next state


        >>> Suspend({"a":1}).resume(lambda s: s)
        Success {'a': 1}

        >>> Success({"a":1}).resume(lambda s: Failed({"error": "Exception!!"}))
        Success {'a': 1}

        >>> Failed({"a":1}).resume(lambda s: Failed({"error": "Exception!!"}))
        Success {'a': 1}

        >>> Suspend({"a":1}).resume(lambda s: Failed({"error": "Exception!!"}))
        Failed {'error': 'Exception!!'}
        """

        next_state = self._fold(Success, Success, Success, Success, Abort, Success, Complete)

        if self.issuspend():
            result = resume_suspend(next_state)
            return result

        return next_state


class Success(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return success(self.s)


class Skipped(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return skipped(self.s)


class Suspend(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return suspend(self.s)


class Waiting(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return waiting(self.s)


class Abort(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return abort(self.s)


class Failed(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return failed(self.s)


class Complete(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return complete(self.s)


_STATUSES = {
    StepStatus.SUCCESS: Success,
    StepStatus.SKIPPED: Skipped,
    StepStatus.SUSPEND: Suspend,
    StepStatus.WAITING: Waiting,
    StepStatus.ABORT: Abort,
    StepStatus.FAILED: Failed,
    StepStatus.COMPLETE: Complete,
}

_NUM_STATUSES = len(_STATUSES)


def cond_bind(log: BoundLogger, state: Dict[str, Any], key: str, as_key: Optional[str] = None) -> BoundLogger:
    """Conditionally (on presence of key) build Structlog context."""
    if as_key is None:
        as_key = key
    if key in state:
        return log.bind(**{as_key: state[key]})
    return log


def log_mutations(old_process_state: State) -> Callable[[State], None]:
    def _log_mutations(new_process_state: State) -> None:
        mutations = {
            k: v for k, v in new_process_state.items() if k not in old_process_state or old_process_state[k] != v
        }
        logger.debug("Step returned a result state.", mutations=mutations)

    return _log_mutations


def errorlogger(error: ErrorDict) -> None:
    logger.error("Workflow returned an error.", **error)


def _exec_steps(steps: StepList, starting_process: Process, dblogstep: StepLogFuncInternal) -> Process:
    """Execute the workflow steps one by one until a Process state other than Success or Skipped is reached."""
    consolelogger = cond_bind(logger, starting_process.unwrap(), "reporter", "created_by")
    process = starting_process

    for step in steps:
        # Check if we need to continue with the process
        if not (process.issuccess() or process.isskipped()):
            break

        consolelogger = consolelogger.bind(step_name=step.name)

        # Debug logging of step information
        mutationlogger = log_mutations(process.unwrap())

        # Execute step
        try:
            engine_status = EngineSettingsTable.query.one()
            if engine_status.global_lock:
                raise RuntimeWarning("Exiting from thread workflow engine is Paused or Pausing")
            step_result_process = process.execute_step(step)
        except RuntimeWarning:
            consolelogger.info(
                "Not executing Step as the workflow engine is Paused. Process will remain in state 'running'"
            )
            return process
        except Exception as e:
            consolelogger.error("An exception occurred while executing the workflow step.")
            step_result_process = Failed(e)

        # write the new process state after the step execution to the database
        # Convert ErrorState to ErrorDict when Failed or Waiting before writing to the database
        # as bare exceptions are not JSON serializable
        result_to_log = step_result_process.on_failed(error_state_to_dict).on_waiting(error_state_to_dict)
        result_to_log.on_success(mutationlogger).on_failed(errorlogger).on_waiting(errorlogger)

        process = dblogstep(step, result_to_log)

        # If database logging failed, the workflow should fail. When it was successful just continue with the
        # result of the executed step.
        consolelogger.debug("Workflow step executed.", process_status=process.status)

    return process


def runwf(pstat: ProcessStat, logstep: StepLogFunc) -> Process:
    """Run workflow optionally adding extra state.

    The extra state is used on resume and to set initial state
    """
    steps = pstat.log

    def _logstep(*x: Any) -> Process:
        return logstep(pstat, *x)

    logger.bind(workflow=pstat.workflow.name)

    def resume_suspend(process: Process) -> Process:
        step = steps.pop(0)
        return _logstep(step, process)

    next_state = pstat.state.resume(resume_suspend)

    return _exec_steps(steps, next_state, _logstep)


def abort_wf(pstat: ProcessStat, logstep: StepLogFunc) -> Process:
    """Abort a suspended workflow."""

    if not pstat.state.iscomplete():
        abort_func = make_step_function(Abort, "User Aborted")

        state = pstat.state.abort()

        return logstep(pstat, abort_func, state)
    else:
        return pstat.state


@_purestep("Start")
def init(state: State) -> Process:
    """Start of workflow."""
    return Success(state)


@_purestep("Done")
def done(state: State) -> Process:
    """End of workflow."""
    return Complete(state)


@_purestep("Abort")
def abort(state: State) -> Process:
    """End of aborted workflow."""
    return Abort(state)


# A `StepList` constructor to be used in reusable step sequences.
begin = StepList()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class IpamStates:
    FREE = 0
    ALLOCATED = 1
    EXPIRED = 2
    PLANNED = 3
    RESERVED = 4
    SUSPEND = 5
    FAILED = 6  # front end code only
    ALLOCATED_SUBNET = 7  # only needed for color coding in frontend

    TRANS = {
        FREE: "Vrij",
        ALLOCATED: "Gealloceerd",
        EXPIRED: "Verlopen",
        PLANNED: "Gepland",
        RESERVED: "Gereserveerd",
        SUSPEND: "Nonactief",
        ALLOCATED_SUBNET: "Gealloceerd subnet",
    }

    @classmethod
    def nl(cls, state: int) -> str:
        return cls.TRANS[state]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
from typing import Callable, Iterable, List, Optional, Sequence, Set, TypeVar, Union

import more_itertools
import structlog

logger = structlog.get_logger(__name__)


def first_available_or_next(values: Iterable[int], start: int = 0) -> int:
    """Return first available value or the next logical value.

    >>> first_available_or_next([0, 1, 3])
    2
    >>> first_available_or_next([0, 1, 2, 3])
    4
    >>> first_available_or_next([1, 2, 3])
    0
    >>> first_available_or_next([])
    0

    >>> first_available_or_next([0, 1, 3], start=11)
    11
    >>> first_available_or_next([], 22)
    22
    >>> first_available_or_next([1, 100, 101], 33)
    33
    >>> first_available_or_next([11, 22, 33, 44, 55], 33)
    34

    Args:
        values: an iterable of integer values.
        start: set starting value.

    Returns:
        First available value or next logical one.

    """
    # +2 -> One +1 to get as many consecutive values up to and including the max value. Another +1 for one extra.
    stop = max(values, default=0) + 2
    if start > stop:
        stop = start + 1
    return min(set(range(start, stop)) - set(values))


def orig(func: Callable) -> Callable:
    """Return the function wrapped by one or more decorators.

    Args:
        func: step function

    Returns:
        Undecorated step function for testing purposes.

    """
    f = func
    while hasattr(f, "__wrapped__"):
        f = f.__wrapped__  # type:ignore

    return f


def join_cs(*args: Union[Iterable[str], str]) -> str:
    """Return comma separated string from one or more comma separated strings or iterables of strings.

    It deals with empty strings and properly inserting comma's.

    See: `test_join_cs` for examples.

    Args:
        args: One or more comma separated strings or iterables that should be joined.

    Returns:
        A comma separated string.

    """

    def to_iterable(value: Union[Iterable[str], str]) -> Iterable[str]:
        if isinstance(value, str):
            return filter(None, value.split(","))
        return value

    return ",".join(itertools.chain(*map(to_iterable, args)))


def expand_ranges(ranges: Sequence[Sequence[int]], inclusive: bool = False) -> List[int]:
    """Expand sequence of range definitions into sorted and deduplicated list of individual values.

    A range definition is either a:

    * one element sequence -> an individual value.
    * two element sequence -> a range of values (either inclusive or exclusive).

    >>> expand_ranges([[1], [2], [10, 12]])
    [1, 2, 10, 11]
    >>> expand_ranges([[1], [2], [10, 12]], inclusive=True)
    [1, 2, 10, 11, 12]
    >>> expand_ranges([[]])
    Traceback (most recent call last):
        ...
    ValueError: Expected 1 or 2 element list for range definition. Got f0 element list instead.

    Resulting list is sorted::

        >>> expand_ranges([[100], [1, 4]], inclusive=True)
        [1, 2, 3, 4, 100]

    Args:
        ranges: sequence of range definitions
        inclusive: are the stop values of the range definition inclusive or exclusive.

    Returns:
        Sorted deduplicated list of individual values.

    Raises:
        ValueError: if range definition is not a one or two element sequence.

    """
    values: Set[int] = set()
    for r in ranges:
        if len(r) == 2:
            values.update(range(r[0], r[1] + (1 if inclusive else 0)))
        elif len(r) == 1:
            values.add(r[0])
        else:
            raise ValueError(f"Expected 1 or 2 element list for range definition. Got f{len(r)} element list instead.")
    return sorted(values)


T = TypeVar("T")


def as_t(value: Optional[T]) -> T:
    """Cast `value` to non-Optional.

    One often needs to assign a value that was typed as being `Optional` to a variable that is typed non-Optional. MyPy
    rightfully takes issue with these assignments (strict Optional checking is default since MyPy 0.600) unless we
    have explicitely determined these values to be not `None`. The most succinct way to do that is using an `assert`
    statement::

        x: Optional[int] = 7
        assert x is not None
        y: int = x

    However that gets tedious pretty fast. One might be inclined to turn off strict Optional checking. However that
    would be a bad decision; None values will percolate through data structures and cause issue at locations far from
    where they originally came from. A better solution would be to fail right where the issue occurred but using a
    somewhat more convenient syntax.

    Some languages such as Kotlin provide the `as` operator:

    .. code-block:: kotlin

        val x: Int? = 7  // ? declaring the Int to be nullable
        val y: Int = x as Int

    That is the inspiration for this function. `t` referring to the type being wrapped in an `Optional`. Hence `as_t`
    meaning `as the non-Optional type`.

    The above Python example now becomes::

        x: Optional[int] = 7
        y: int = as_t(x)

    `as_t` checks whether te value passed to it is not `None`, satisfying MyPy. If it happens to be `None` it raises a
    `ValueError`, satisfying our requirement to fail at the location where we require the value to be not None and not
    somewhere far down the code path.

    Args:
        value: `Optional` value to be casted to non-Optional

    Returns:
        non-Optional value.

    Raises:
        ValueError: in case `value` is `None`

    """
    if value is None:
        raise ValueError("Trying to cast a value to non-Optional type failed due to value being None.")
    return value


def ireplace(iterable: Iterable[T], old: T, new: T) -> Iterable[T]:
    """Replace one or more occurrences of a specific value in an iterable with another value.

    The 'i' prefix indicates 'iterable' and is there to distinguish it from other similar functions.

    >>> list(ireplace(["1-10", "", "22"], "", "0"))
    ['1-10', '0', '22']

    Args:
        iterable: The iterable that needs to have a specific value replaced for all its occurrences.
        old: The value in the iterable to replace.
        new: The value to replace `old` with.

    Returns:
        A new iterable with `old` values replaced by `new` values

    """
    yield from more_itertools.replace(iterable, lambda v: v == old, [new])


def to_ranges(i: Iterable[int]) -> Iterable[range]:
    """Convert a sorted iterable of ints to an iterable of range objects.

    IMPORTANT: the iterable passed in should be sorted and not contain duplicate elements.

    Examples::
        >>> list(to_ranges([2, 3, 4, 5, 7, 8, 9, 45, 46, 47, 49, 51, 53, 54, 55, 56, 57, 58, 59, 60, 61]))
        [range(2, 6), range(7, 10), range(45, 48), range(49, 50), range(51, 52), range(53, 62)]

    Args:
        i: sorted iterable

    Yields:
        range object for each consecutive set of integers

    """
    # The trick here is the key function (the lambda one) that calculates the difference between an element of the
    # iterable `i` and its corresponding enumeration value. For consecutive values in the iterable, this difference
    # will be the same! All these values (those with the same difference) are grouped by the `groupby` function. We
    # return the first and last element to construct a `range` object
    for _, g in itertools.groupby(enumerate(i), lambda t: t[1] - t[0]):
        group = list(g)
        yield range(group[0][1], group[-1][1] + 1)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import List, Optional
from uuid import UUID

from pydantic import Extra

from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.schemas.product import ProductBaseSchema
from orchestrator.schemas.product_block import ProductBlockSchema
from orchestrator.schemas.resource_type import ResourceTypeSchema
from orchestrator.schemas.subscription_descriptions import SubscriptionDescriptionSchema
from orchestrator.types import SubscriptionLifecycle, strEnum


class PortMode(strEnum):
    """Valid port modes."""

    TAGGED = "tagged"
    UNTAGGED = "untagged"
    LINKMEMBER = "link_member"


class SubscriptionRelationSchema(OrchestratorBaseModel):
    domain_model_attr: str
    child_id: UUID
    parent_id: UUID
    order_id: int

    class Config:
        orm_mode = True


class SubscriptionInstanceValueBaseSchema(OrchestratorBaseModel):
    resource_type_id: UUID
    subscription_instance_id: UUID
    subscription_instance_value_id: UUID
    value: str
    resource_type: ResourceTypeSchema

    class Config:
        orm_mode = True


class SubscriptionInstanceBase(OrchestratorBaseModel):
    label: Optional[str]
    subscription_id: UUID
    product_block_id: UUID
    subscription_instance_id: UUID
    values: List[SubscriptionInstanceValueBaseSchema]
    parent_relations: List[SubscriptionRelationSchema]
    children_relations: List[SubscriptionRelationSchema]
    product_block: ProductBlockSchema

    class Config:
        orm_mode = True


class SubscriptionBaseSchema(OrchestratorBaseModel):
    subscription_id: Optional[UUID]
    start_date: Optional[datetime]
    description: str
    status: SubscriptionLifecycle
    product_id: Optional[UUID]
    customer_id: UUID
    insync: bool
    note: Optional[str]


class SubscriptionSchema(SubscriptionBaseSchema):
    name: Optional[str]
    subscription_id: UUID
    end_date: Optional[datetime]
    product: Optional[ProductBaseSchema]
    customer_descriptions: Optional[List[SubscriptionDescriptionSchema]]
    tag: Optional[str]

    class Config:
        orm_mode = True


class SubscriptionIdSchema(OrchestratorBaseModel):
    subscription_id: UUID


class SubscriptionWithPortAttrSchema(SubscriptionSchema):
    port_mode: Optional[str]


class SubscriptionDetailsSchema(SubscriptionWithPortAttrSchema):
    instances: List[SubscriptionInstanceBase]
    customer_descriptions: List[SubscriptionDescriptionSchema]


class SubscriptionDomainModelSchema(SubscriptionSchema):
    customer_descriptions: List[SubscriptionDescriptionSchema]

    class Config:
        extra = Extra.allow



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def remove_redundant_ws(line: str) -> str:
    """Remove redundant white space from a line.

    Redundant being multiple spaces where only one is needed.

    >>> remove_redundant_ws(" a  b  c ")
    'a b c'

    >>> remove_redundant_ws("a b c")
    'a b c'

    >>> remove_redundant_ws("   ")
    ''

    Args:
        line: the string to remove redundant white space from

    Returns:
        Cleaned up string with only one space between textual elements.

    """
    return " ".join(line.split())



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orchestrator.db import db
from orchestrator.forms import FormPage
from orchestrator.forms.validators import LongText
from orchestrator.services import subscriptions
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, UUIDstr
from orchestrator.workflow import StepList, done, init, workflow
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import wrap_modify_initial_input_form


def initial_input_form(subscription_id: UUIDstr) -> FormGenerator:
    subscription = subscriptions.get_subscription(subscription_id)
    old_note = subscription.note

    class ModifyNoteForm(FormPage):
        note: LongText = old_note

    user_input = yield ModifyNoteForm
    subscription.note = user_input.note
    db.session.add(subscription)
    return {"old_note": old_note, "note": user_input.note}


@workflow("Modify Note", initial_input_form=wrap_modify_initial_input_form(initial_input_form), target=Target.MODIFY)
def modify_note() -> StepList:
    return init >> store_process_subscription(Target.MODIFY) >> done



from copy import deepcopy
from functools import reduce
from typing import List, NoReturn, Tuple
from unittest import mock
from uuid import UUID, uuid4

import pytest
from nwastdlib import const

from orchestrator.config.assignee import Assignee
from orchestrator.forms import FormPage
from orchestrator.services.processes import SYSTEM_USER
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.workflow import Abort, Failed, Process, ProcessStat, Skipped, Success, Suspend, Waiting
from orchestrator.workflow import _purestep as purestep
from orchestrator.workflow import (
    abort,
    abort_wf,
    begin,
    conditional,
    done,
    focussteps,
    init,
    inputstep,
    retrystep,
    runwf,
    step,
    workflow,
)
from test.unit_tests.workflows import (
    WorkflowInstanceForTests,
    assert_aborted,
    assert_complete,
    assert_failed,
    assert_state,
    assert_success,
    assert_suspended,
    assert_waiting,
    extract_error,
    resume_workflow,
    run_workflow,
)


@step("Step 1")
def step1():
    return {"steps": [1]}


@step("Step 2")
def step2(steps):
    return {"steps": [*steps, 2]}


@step("Step 3")
def step3(steps):
    return {"steps": [*steps, 3]}


@inputstep("Input Name", assignee=Assignee.SYSTEM)
def user_action():
    class Form(FormPage):
        name: str

    return Form


@retrystep("Waiting step")
def soft_fail():
    raise ValueError("Failure Message")


@step("Fail")
def fail():
    raise ValueError("Failure Message")


@workflow("Sample workflow")
def sample_workflow():
    return begin >> step1 >> step2 >> step3


def test_process_state_assertions():
    process_success = Success({"done": True})
    assert_success(process_success)
    with pytest.raises(AssertionError):
        assert_failed(process_success)

    process_failed_with_message = Failed("Failure message")

    assert_failed(process_failed_with_message)

    with pytest.raises(AssertionError):
        assert_success(process_failed_with_message)

    process_failed_with_exception = Failed(ValueError("A is not B"))

    assert_failed(process_failed_with_exception)

    with pytest.raises(ValueError):
        # When a Failed or Waiting state is encountered with an exception it will be reraised
        assert_success(process_failed_with_exception)


def create_new_process_stat(workflow, initial_state):
    return ProcessStat(
        pid=str(uuid4()), workflow=workflow, state=Success(initial_state), log=workflow.steps, current_user=SYSTEM_USER
    )


def test_exec_through_all_steps():
    log = []

    pstat = create_new_process_stat(sample_workflow, {})
    result = runwf(pstat, store(log))
    assert_success(result)
    assert_state(result, {"steps": [1, 2, 3]})


def test_store_all_steps():
    log = []

    pstat = create_new_process_stat(sample_workflow, {})
    runwf(pstat, store(log))

    assert [
        ("Step 1", Success({"steps": [1]})),
        ("Step 2", Success({"steps": [1, 2]})),
        ("Step 3", Success({"steps": [1, 2, 3]})),
    ] == log


def test_recover():
    log = []

    p = ProcessStat(
        pid=1,
        workflow=sample_workflow,
        state=Success({"steps": [4]}),
        log=sample_workflow.steps[1:],
        current_user="john.doe",
    )
    result = runwf(p, store(log))
    assert_success(result)
    assert_state(result, {"steps": [4, 2, 3]})


def test_waiting():
    wf = workflow("Workflow with soft fail")(lambda: begin >> step1 >> soft_fail >> step2)

    log = []

    pstat = create_new_process_stat(wf, {})
    result = runwf(pstat, store(log))

    assert_waiting(result)
    assert extract_error(result) == "Failure Message"

    assert [
        ("Step 1", Success({"steps": [1]})),
        ("Waiting step", Waiting({"class": "ValueError", "error": "Failure Message", "traceback": mock.ANY})),
    ] == log


def test_resume_waiting_workflow():
    hack = {"error": True}

    @retrystep("Waiting step")
    def soft_fail():
        if hack["error"]:
            raise ValueError("error")
        else:
            return {"some_key": True}

    wf = workflow("Workflow with soft fail")(lambda: begin >> step1 >> soft_fail >> step2)

    log = []

    state = Waiting({"steps": [1]})

    hack["error"] = False
    p = ProcessStat(pid=1, workflow=wf, state=state, log=wf.steps[1:], current_user="john.doe")
    result = runwf(p, logstep=store(log))

    assert_success(result)
    assert [
        ("Waiting step", Success({"steps": [1], "some_key": True})),
        ("Step 2", Success({"steps": [1, 2], "some_key": True})),
    ] == log


def test_suspend():
    wf = workflow("Workflow with user interaction")(lambda: begin >> step1 >> user_action >> step2)

    log = []

    pstat = create_new_process_stat(wf, {})
    result = runwf(pstat, store(log))

    assert_suspended(result)
    assert [("Step 1", Success({"steps": [1]})), ("Input Name", Suspend({"steps": [1]}))] == log


def test_resume_suspended_workflow():
    wf = workflow("Workflow with user interaction")(lambda: begin >> step1 >> user_action >> step2)

    log = []

    p = ProcessStat(
        pid=1,
        workflow=wf,
        state=Suspend({"steps": [1], "name": "Jane Doe"}),
        log=wf.steps[1:],
        current_user="john.doe",
    )
    result = runwf(p, logstep=store(log))

    assert_success(result)
    assert result == Success({"steps": [1, 2], "name": "Jane Doe"})
    assert [
        ("Input Name", Success({"steps": [1], "name": "Jane Doe"})),
        ("Step 2", Success({"steps": [1, 2], "name": "Jane Doe"})),
    ] == log


def test_abort():
    wf = workflow("Aborting workflow")(lambda: init >> abort)

    log = []

    pstat = create_new_process_stat(wf, {"name": "aborting"})
    result = runwf(pstat, store(log))
    assert_aborted(result)
    assert_state(result, {"name": "aborting"})


def test_failed_step():
    wf = workflow("Failing workflow")(lambda: init >> fail)

    log = []

    pstat = create_new_process_stat(wf, {"name": "init-state"})
    result = runwf(pstat, store(log))
    assert_failed(result)
    assert extract_error(result) == "Failure Message"
    assert [
        ("Start", Success({"name": "init-state"})),
        ("Fail", Failed({"class": "ValueError", "error": "Failure Message", "traceback": mock.ANY})),
    ] == log


def test_failed_log_step():
    wf = workflow("Failing workflow")(lambda: init >> done)

    def failing_store(stat, step, state):
        return Failed("Failure Message")

    pstat = create_new_process_stat(wf, {"name": "init-state"})
    result = runwf(pstat, failing_store)
    assert_failed(result)
    assert extract_error(result) == "Failure Message"


def test_exception_log_step():
    wf = workflow("Failing workflow")(lambda: init >> done)

    def failing_store(stat, step, state):
        raise Exception("Failing store error")

    with pytest.raises(Exception) as exc_info:
        pstat = create_new_process_stat(wf, {"name": "init-state"})
        runwf(pstat, failing_store)
    assert "Failing store error" in str(exc_info.value)


def test_complete():
    wf = workflow("WF")(lambda: init >> done)

    log = []
    pstat = create_new_process_stat(wf, {"name": "completion"})
    result = runwf(pstat, store(log))
    assert_complete(result)
    assert_state(result, {"name": "completion"})


def test_abort_workflow():
    wf = workflow("Workflow with user interaction")(lambda: begin >> step1 >> user_action)

    log = []
    state = {"steps": [1]}
    pstat = ProcessStat(pid=1, workflow=wf, state=Success(state), log=wf.steps[1:], current_user="john.doe")

    result = abort_wf(pstat, store(log))

    assert Abort({"steps": [1]}) == result
    assert log == [("User Aborted", Abort(state))]


def test_focus_state():
    @step("Step that works on substate")
    def substep():
        return {"result": "substep"}

    subwf = focussteps("sub")
    wf = workflow("Workflow with sub workflow that focuses on sub state")(lambda: subwf(substep) >> done)

    log = []
    pstat = create_new_process_stat(wf, {"sub": {}})
    result = runwf(pstat, store(log))
    assert_complete(result)
    assert_state(result, {"sub": {"result": "substep"}})

    # Test on empty key
    subwf = focussteps("sub")
    wf = workflow("Workflow with sub workflow that focuses on sub state")(lambda: subwf(substep) >> done)

    log = []
    pstat = create_new_process_stat(wf, {})
    result = runwf(pstat, store(log))
    assert_complete(result)
    assert_state(result, {"sub": {"result": "substep"}})


def test_error_in_focus_state():
    @step("Step that works on substate")
    def substep():
        raise ValueError("Error")

    subwf = focussteps("sub")
    wf = workflow("Workflow with sub workflow that focuses on sub state")(lambda: subwf(substep) >> done)

    log = []
    pstat = create_new_process_stat(wf, {"sub": {}})
    result = runwf(pstat, store(log))
    assert_failed(result)
    assert extract_error(result) == "Error"


def test_input_in_substate() -> None:
    @inputstep("Input Name", assignee=Assignee.SYSTEM)
    def input_action(state: State) -> FormGenerator:
        class SubForm(FormPage):
            a: int

        class TestForm(FormPage):
            sub: SubForm

        user_input = yield TestForm

        return user_input.dict()

    wf = workflow("Workflow with user interaction")(
        lambda: begin >> input_action >> purestep("process inputs")(Success)
    )

    log: List[Tuple[str, Process]] = []
    pid = uuid4()
    p = ProcessStat(
        pid=pid, workflow=wf, state=Suspend({"sub": {"a": 1, "b": 2}}), log=wf.steps[1:], current_user="john.doe"
    )
    result = runwf(p, logstep=store(log))

    assert_success(result)
    assert_state(result, {"sub": {"a": 1, "b": 2}})


def test_skip_step():
    wf = workflow("Workflow with skipped step")(lambda: init >> purestep("Skipped")(Skipped) >> done)

    log = []
    pstat = create_new_process_stat(wf, {})
    result = runwf(pstat, store(log))
    assert_complete(result)

    skipped = [x[1] for x in log if x[0] == "Skipped"]
    assert skipped[0].isskipped()


def test_conditionally_skip_a_step():
    @step("Inc N")
    def inc_n(n=0):
        return {"n": n + 1}

    limit_to_10 = conditional(lambda s: s.get("n", 0) < 10)

    incs = [limit_to_10(inc_n) for _ in range(0, 25)]
    wf = workflow("Limit the number of increments")(lambda: init >> reduce(lambda acc, e: acc >> e, incs) >> done)

    log = []

    pstat = create_new_process_stat(wf, {})
    result = runwf(pstat, store(log))
    assert_complete(result)
    # from ipdb import set_trace; set_trace()
    assert_state(result, {"n": 10})
    assert len([x for x in log if x[1].isskipped()]) == 15, "15 steps should be skipped"


def store(log):
    def _store(_, step, state):
        log.append((step.name, state))
        return state

    return _store


def test_failing_inputstep_with_form_state_params() -> None:
    @inputstep("Modify")
    def modify(subscription_id: UUIDstr) -> NoReturn:
        raise Exception("Something went wrong")

    class Form(FormPage):
        subscription_id: UUID

    @workflow("Workflow", initial_input_form=const(Form))
    def inject_args_test_workflow():
        return init >> modify >> done

    WorkflowInstanceForTests(inject_args_test_workflow, "inject_args_test_workflow")

    init_state = {"subscription_id": uuid4()}

    result, process, step_log = run_workflow("inject_args_test_workflow", init_state)
    assert_suspended(result)
    step_log_copy = deepcopy(step_log)

    with pytest.raises(Exception) as e:
        resume_workflow(process, step_log, {})

    assert "Something went wrong" in str(e.value)
    assert step_log_copy == step_log  # No steps have been logged because of the error


def test_update_wrapper():
    @step("Whatever")
    def whatever():
        """Do whatever."""
        pass

    assert whatever.__doc__ == "Do whatever."



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orchestrator.workflow import StepList, workflow


# This workflow has been made to create the initial import process for a SN7 subscription
# it does not do anything but is needed for the correct showing in the GUI.
@workflow("Dummy workflow to replace removed workflows")
def removed_workflow() -> StepList:
    return StepList()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from datetime import datetime
from itertools import groupby, zip_longest
from operator import attrgetter
from typing import Any, Callable, ClassVar, Dict, List, Optional, Set, Type, TypeVar, Union
from uuid import UUID, uuid4

import structlog
from more_itertools import flatten, only
from pydantic import BaseModel, Field, ValidationError
from pydantic.main import ModelMetaclass
from pydantic.types import ConstrainedList
from sqlalchemy import and_
from sqlalchemy.orm import selectinload

from orchestrator.db import (
    ProductBlockTable,
    ProductTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.domain.lifecycle import ProductLifecycle, lookup_specialized_type, register_specialized_type
from orchestrator.types import State, SubscriptionLifecycle, UUIDstr, is_list_type, is_of_type, is_optional_type
from orchestrator.utils.docs import make_product_block_docstring, make_subscription_model_docstring

logger = structlog.get_logger(__name__)


def _is_constrained_list_type(type: Type) -> bool:
    # subclass on typing.List throws exception and there is no good way to test for this
    try:
        is_constrained_list = issubclass(type, ConstrainedList)
    except Exception:

        # Strip generic arguments, it still might be a subclass
        if hasattr(type, "__origin__"):
            return _is_constrained_list_type(type.__origin__)
        else:
            return False

    return is_constrained_list


T = TypeVar("T")
S = TypeVar("S", bound="SubscriptionModel")
B = TypeVar("B", bound="ProductBlockModel")


class DomainModel(BaseModel):
    """Base class for domain models.

    Contains all common Product block/Subscription instance code
    """

    class Config:
        validate_assignment = True
        validate_all = True
        arbitrary_types_allowed = True

    __base_type__: ClassVar[Type["DomainModel"]]
    _product_block_fields_: ClassVar[Dict[str, Type]]
    _non_product_block_fields_: ClassVar[Dict[str, Type]]

    def __init_subclass__(
        cls, *args: Any, lifecycle: Optional[List[SubscriptionLifecycle]] = None, **kwargs: Any
    ) -> None:
        super().__init_subclass__()
        cls._find_special_fields()

        if kwargs.keys():
            logger.warning(
                "Unexpected keyword arguments in domain model class", class_name=cls.__name__, kwargs=kwargs.keys()
            )

        # Check if child subscription instance models conform to the same lifecycle
        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            if is_list_type(product_block_field_type) or is_optional_type(product_block_field_type):
                product_block_field_type = product_block_field_type.__args__[0]

            if lifecycle:
                for lifecycle_status in lifecycle:
                    if lookup_specialized_type(product_block_field_type, lifecycle_status) != product_block_field_type:
                        raise AssertionError(
                            f"The lifecycle status of the type for the field: {product_block_field_name} is not suitable for the lifecycle status of this model"
                        )

    @classmethod
    def _find_special_fields(cls: Type) -> None:
        """Make and store a list of resource_type fields and product block fields."""
        cls._non_product_block_fields_ = {}
        cls._product_block_fields_ = {}

        for field_name, field_type in cls.__annotations__.items():
            try:
                is_product_block_field = (
                    is_list_type(field_type, DomainModel)
                    or is_optional_type(field_type, DomainModel)
                    or is_of_type(field_type, DomainModel)
                )
            except TypeError:
                # issubclass does not work on typing types
                is_product_block_field = False

            # We only want fields that are on this class and not on the parent
            if is_product_block_field:
                cls._product_block_fields_[field_name] = field_type
            else:
                cls._non_product_block_fields_[field_name] = field_type

    @classmethod
    def _init_instances(
        cls, skip_keys: Optional[List[str]] = None
    ) -> Dict[str, Union[List["ProductBlockModel"], "ProductBlockModel"]]:
        """Initialize default subscription instances.

        When a new domain model is created that is not loaded from an existing subscription.
        We also create all subscription instances for it. This function does that.

        Args:
            skip_keys: list of fields on the class to skip when creating dummy instances.

        Returns:
            A dict with instances to pass to the new model

        """
        if skip_keys is None:
            skip_keys = []

        instances: Dict[str, Union[List[ProductBlockModel], ProductBlockModel]] = {}
        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            if product_block_field_name in skip_keys:
                continue

            if is_list_type(product_block_field_type):
                if _is_constrained_list_type(product_block_field_type):
                    product_block_model = product_block_field_type.__args__[0]
                    default_value = product_block_field_type()
                    # if constrainedlist has minimum, return that minimum else empty list
                    if product_block_field_type.min_items:
                        logger.debug("creating min_items", type=product_block_field_type)
                        for _ in range(product_block_field_type.min_items):
                            default_value.append(product_block_model.new())
                else:
                    # a list field of ProductBlockModels without limits gets an empty list
                    default_value = []
            elif is_optional_type(product_block_field_type, ProductBlockModel):
                default_value = None
            elif issubclass(product_block_field_type, ProductBlockModel):
                product_block_model = product_block_field_type
                # Scalar field of a ProductBlockModel expects 1 instance
                default_value = product_block_model.new()
            else:
                raise AssertionError("ProductTable block field has an unsupported type")
            instances[product_block_field_name] = default_value
        return instances

    @classmethod
    def _load_instances(
        cls, db_instances: List[SubscriptionInstanceTable], status: SubscriptionLifecycle
    ) -> Dict[str, Union[Optional["ProductBlockModel"], List["ProductBlockModel"]]]:
        """Load subscription instances for this domain model.

        When a new domain model is loaded from an existing subscription we also load all
        subscription instances for it. This function does that.

        Args:
            db_instances: list of database models to load from
            status: SubscriptionLifecycle of subscription to check if models match

        Returns:
            A dict with instances to pass to the new model

        """

        instances: Dict[str, Union[Optional[ProductBlockModel], List[ProductBlockModel]]] = {}

        def keyfunc(i: SubscriptionInstanceTable) -> str:
            return i.product_block.name

        sorted_instances = sorted(db_instances, key=keyfunc)
        grouped_instances = {k: list(g) for k, g in groupby(sorted_instances, keyfunc)}

        def match_domain_model_attr_if_possible(field_name: str) -> Callable:
            def domain_filter(instance: SubscriptionInstanceTable) -> bool:
                """
                Match domain model attributes.

                This helper is necessary to filter through all relations in a subscription. Not all subscriptions have a
                domain model attribute that is set as it is not always necessary. However when it is set, it is necessary
                to filter through instances depending on that attribute.

                Args:
                    instance: child instance

                Returns:
                    Boolean of match.

                """
                attr_names = {
                    relation.domain_model_attr for relation in instance.parent_relations if relation.domain_model_attr
                }

                # We can assume true is no domain_model_attr is set.
                return not attr_names or field_name in attr_names

            return domain_filter

        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            filter_func = match_domain_model_attr_if_possible(product_block_field_name)
            if is_list_type(product_block_field_type):
                if product_block_field_name not in instances:
                    if _is_constrained_list_type(product_block_field_type):
                        product_block_model_list = product_block_field_type()
                    else:
                        product_block_model_list = []
                else:
                    product_block_model_list = instances[product_block_field_name]

                product_block_model = product_block_field_type.__args__[0]
                instance_list: List[SubscriptionInstanceTable] = list(
                    filter(
                        filter_func, flatten(grouped_instances.get(name, []) for name in product_block_model.__names__)
                    )
                )
                product_block_model_list.extend(
                    product_block_model.from_db(subscription_instance=instance, status=status)
                    for instance in instance_list
                )

                instances[product_block_field_name] = product_block_model_list
            else:
                product_block_model = product_block_field_type
                if is_optional_type(product_block_field_type):
                    product_block_model = product_block_model.__args__[0]

                instance = only(
                    list(
                        filter(
                            filter_func,
                            flatten(grouped_instances.get(name, []) for name in product_block_model.__names__),
                        )
                    )
                )
                if is_optional_type(product_block_field_type) and instance is None:
                    instances[product_block_field_name] = None
                elif not is_optional_type(product_block_field_type) and instance is None:
                    raise ValueError("Required subscription instance is missing in database")
                else:
                    instances[product_block_field_name] = product_block_model.from_db(
                        subscription_instance=instance, status=status
                    )

        return instances

    def _save_instances(
        self, subscription_id: UUID, instances: Dict[UUID, SubscriptionInstanceTable], status: SubscriptionLifecycle
    ) -> Dict[str, Union[SubscriptionInstanceTable, List[SubscriptionInstanceTable]]]:
        """Save subscription instances for this domain model.

        When a domain model is saved to the database we need to save all child subscription instances for it.

        Args:
            subscription_id: The subscription id
            instances: list of existing database models to update
            status: SubscriptionLifecycle of subscription to check if models match

        Returns:
            A list with instances to save

        """
        saved_instances: Dict[str, Union[SubscriptionInstanceTable, List[SubscriptionInstanceTable]]] = {}
        for product_block_field, product_block_field_type in self._product_block_fields_.items():
            product_block_models = getattr(self, product_block_field)
            if is_list_type(product_block_field_type):
                saved_instances[product_block_field] = []
                for product_block_model in product_block_models:
                    saved_instances[product_block_field].extend(  # type: ignore
                        product_block_model.save(
                            subscription_id=subscription_id, subscription_instances=instances, status=status
                        )
                    )
            elif is_optional_type(product_block_field_type) and product_block_models is None:
                pass
            else:
                saved_instances[product_block_field] = product_block_models.save(
                    subscription_id=subscription_id, subscription_instances=instances, status=status
                )

        return saved_instances


class ProductBlockModelMeta(ModelMetaclass):
    """Metaclass used to create product block instances.

    This metaclass is used to make sure the class contains product block metadata.

    This metaclass should not be used directly in the class definition. Instead a new product block model should inherit
    from ProductBlockModel which has this metaclass defined.

    You can find some examples in: :ref:`domain-models`
    """

    __names__: List[str]
    name: str
    product_block_id: UUID
    description: str
    tag: str

    def __call__(self, *args: Any, **kwargs: Any) -> B:

        # Would have been nice to do this in __init_subclass__ but that runs outside the app context so we cant access the db
        # So now we do it just before we instantiate the instance
        if not hasattr(self, "product_block_id"):
            product_block = ProductBlockTable.query.filter(ProductBlockTable.name == self.name).one()
            self.product_block_id = product_block.product_block_id
            self.description = product_block.description
            self.tag = product_block.tag

        kwargs["name"] = self.name

        return super().__call__(*args, **kwargs)


class ProductBlockModel(DomainModel, metaclass=ProductBlockModelMeta):
    """Base class for all product block models.

    ProductTable Blocks are represented as dataclasses with pydantic runtime validation.

    Different stages of a subscription lifecycle could require different product block definition. Mainly to support
    mandatory fields when a subscription is active. To support this a lifecycle specific product block definition can
    be created by subclassing the generic product block with keyword argument 'lifecycle' and overriding its fields.

    All product blocks are related to a database ProductBlockTable object through the `product_block_name` that is given
    as class keyword argument.

    Define a product block:
    >>> class BlockInactive(ProductBlockModel, product_block_name="Virtual Circuit"):
    ...    int_field: Optional[int] = None
    ...    str_field: Optional[str] = None

    >>> class Block(BlockInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    ...    int_field: int
    ...    str_field: str

    This example defines a product_block with two different contraints based on lifecycle. `Block` is valid only for `ACTIVE`
    And `BlockInactive` for all other states.
    `product_block_name` must be defined on the base class and need not to be defined on the others

    Create a new empty product block
    >>> example1 = BlockInactive()  # doctest:+SKIP

    Create a new instance based on a dict in the state:
    >>> example2 = BlockInactive(**state)  # doctest:+SKIP

    To retrieve a ProductBlockModel from the database.:
    >>> BlockInactive.from_db(subscription_instance_id)  # doctest:+SKIP
    """

    __names__: ClassVar[List[str]]
    product_block_id: ClassVar[UUID]
    description: ClassVar[str]
    tag: ClassVar[str]

    name: str  # Product block name. This needs to be an instance var because its part of the API (we expose it to the frontend)
    subscription_instance_id: UUID = Field(default_factory=uuid4)
    label: Optional[str] = None

    def __init_subclass__(
        cls,
        *,
        product_block_name: Optional[str] = None,
        product_block_names: Optional[List[str]] = None,
        lifecycle: Optional[List[SubscriptionLifecycle]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(lifecycle=lifecycle, **kwargs)

        if product_block_name is not None:
            cls.name = product_block_name
            cls.__base_type__ = cls
            cls.__names__ = product_block_names or [cls.name]

        if product_block_name is not None or lifecycle is not None:
            register_specialized_type(cls, lifecycle)

        cls.__doc__ = make_product_block_docstring(cls, lifecycle)

    @classmethod
    def new(cls: Type[B], **kwargs: Any) -> B:
        """Create a new empty product block.

        We need to use this instead of the normal constructor because that assumes you pass in
        all required values. That is cumbersome since that means creating a tree of product blocks.

        This is similar to `from_product_id()`
        """
        sub_instances = cls._init_instances(list(kwargs.keys()))

        return cls(**sub_instances, **kwargs)  # type: ignore

    @classmethod
    def _load_instances_values(cls, instance_values: List[SubscriptionInstanceValueTable]) -> Dict[str, str]:
        """Load non product block fields (instance values).

        Args:
            instance_values: List of instance values from database

        Returns:
            Dict of fields to use for constructor

        """
        instance_values_dict: State = {}
        list_field_names = []

        # Set default values
        for field_name, field_type in cls._non_product_block_fields_.items():
            # Ensure that empty lists are handled OK
            if is_list_type(field_type):
                instance_values_dict[field_name] = []
                list_field_names.append(field_name)

        for siv in instance_values:
            # check the type of the siv in the instance and act accordingly: only lists and scalar values supported
            resource_type_name = siv.resource_type.resource_type
            if is_list_type(cls._non_product_block_fields_[resource_type_name]):
                instance_values_dict[resource_type_name].append(siv.value)
            else:
                instance_values_dict[resource_type_name] = siv.value

        # Make sure values are sorted. This already happens when they come from the db.
        # However newly created SubscriptionInstances might not have the correct order
        for field_name in list_field_names:
            instance_values_dict[field_name] = sorted(instance_values_dict[field_name])

        return instance_values_dict

    @classmethod
    def from_db(
        cls: Type[B],
        subscription_instance_id: Optional[UUID] = None,
        subscription_instance: Optional[SubscriptionInstanceTable] = None,
        status: Optional[SubscriptionLifecycle] = None,
    ) -> B:
        """Create a product block based on a subscription instance from the database.

        This function is similar to `from_subscription()`

        >>> subscription_instance_id = KNOWN_UUID_IN_DB  # doctest:+SKIP
        >>> si_from_db = db.SubscriptionInstanceTable.query.get(subscription_instance_id)  # doctest:+SKIP
        >>> example3 = ProductBlockModel.from_db(subscription_instance=si_from_db)  # doctest:+SKIP
        >>> example4 = ProductBlockModel.from_db(subscription_instance_id=subscription_instance_id)  # doctest:+SKIP
        """
        # Fill values from actual subscription
        if subscription_instance_id:
            subscription_instance = SubscriptionInstanceTable.query.get(subscription_instance_id)
        if subscription_instance:
            subscription_instance_id = subscription_instance.subscription_instance_id
        assert subscription_instance_id
        assert subscription_instance

        if not status:
            status = SubscriptionLifecycle(subscription_instance.subscription.status)  # type: ignore
        if not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for lifecycle {status}")

        label = subscription_instance.label

        instance_values = cls._load_instances_values(subscription_instance.values)
        sub_instances = cls._load_instances(subscription_instance.children, status)

        try:
            return cls(  # type: ignore
                subscription_instance_id=subscription_instance_id,
                label=label,
                **instance_values,
                **sub_instances,
            )
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database",
                loaded_instance_values=instance_values,
                loaded_sub_instances=sub_instances,
            )
            raise

    def _save_instance_values(
        self, product_block: ProductBlockTable, current_values: List[SubscriptionInstanceValueTable]
    ) -> List[SubscriptionInstanceValueTable]:
        """Save non product block fields (instance values).

        Returns:
            List of database instances values to save

        """
        resource_types = {rt.resource_type: rt for rt in product_block.resource_types}
        current_values_dict: Dict[str, List[SubscriptionInstanceValueTable]] = defaultdict(list)
        for siv in current_values:
            current_values_dict[siv.resource_type.resource_type].append(siv)

        subscription_instance_values = []
        for field_name, field_type in self._non_product_block_fields_.items():
            assert (
                field_name in resource_types
            ), f"Domain model {self.__class__} does not match the ProductBlockTable {product_block.name}, missing: {field_name} {resource_types}"

            resource_type = resource_types[field_name]
            value = getattr(self, field_name)
            if value is None:
                continue
            if is_list_type(field_type):
                for val, siv in zip_longest(value, current_values_dict[field_name]):
                    if val is not None:
                        if siv:
                            siv.value = str(val)
                            subscription_instance_values.append(siv)
                        else:
                            subscription_instance_values.append(
                                SubscriptionInstanceValueTable(resource_type=resource_type, value=str(val))
                            )
            else:
                if field_name in current_values_dict:
                    current_value = current_values_dict[field_name][0]
                    current_value.value = str(value)
                    subscription_instance_values.append(current_value)
                else:
                    subscription_instance_values.append(
                        SubscriptionInstanceValueTable(resource_type=resource_type, value=str(value))
                    )
        return subscription_instance_values

    def _set_instance_domain_model_attrs(
        self,
        subscription_instance_mapping: Dict[str, Union[SubscriptionInstanceTable, List[SubscriptionInstanceTable]]],
    ) -> None:
        """
        Save the domain model attribute to the database.

        This function iterates through the subscription instances and stores the domain model attribute in the
        hierarchy relationship.

        Args:
            subscription_instance_mapping: a mapping of the domain model attribute a underlying instances

        Returns:
            None

        """
        # Set the domain_model_attrs in the database
        for domain_model_attr, instances in subscription_instance_mapping.items():
            if isinstance(instances, list):
                for instance in instances:
                    relation = SubscriptionInstanceRelationTable.query.filter(
                        and_(
                            SubscriptionInstanceRelationTable.parent_id == self.subscription_instance_id,
                            SubscriptionInstanceRelationTable.child_id == instance.subscription_instance_id,
                        )
                    ).one_or_none()

                    # Skip the None case if no relation is found.
                    if relation:
                        relation.domain_model_attr = domain_model_attr
            else:
                relation = SubscriptionInstanceRelationTable.query.filter(
                    and_(
                        SubscriptionInstanceRelationTable.parent_id == self.subscription_instance_id,
                        SubscriptionInstanceRelationTable.child_id == instances.subscription_instance_id,
                    )
                ).one_or_none()

                # Some subscription instances do not have a parent, so skip the None case
                if relation:
                    relation.domain_model_attr = domain_model_attr

    def save(
        self,
        *,
        status: SubscriptionLifecycle,
        subscription_id: Optional[UUID] = None,
        subscription_instances: Optional[Dict[UUID, SubscriptionInstanceTable]] = None,
    ) -> List[SubscriptionInstanceTable]:
        """Save the current model instance to the database.

        This means saving the whole tree of subscription instances and seperately saving all instance values for this instance.

        Args:
            status: current SubscriptionLifecycle to check if all constraints match
            subscription_id: Optional subscription id needed if this is a new model
            subscription_instances: List of existing subscription instances to update

        Returns:
            List of saved instances

        """
        if len(self.__names__) > 1:
            raise TypeError(
                "You're calling save() from within one of the Generic/Base ProductTable Blocks: which are read-only by design"
            )

        # Make sure we have a valid subscription instance database model
        if subscription_instances is None:
            subscription_instances = {}

        subscription_instance = None
        if subscription_instances and self.subscription_instance_id in subscription_instances:
            subscription_instance = subscription_instances[self.subscription_instance_id]
            subscription_id = subscription_instance.subscription_id

        if subscription_instance is None and subscription_id:
            subscription_instance = SubscriptionInstanceTable(
                subscription_instance_id=self.subscription_instance_id,
                subscription_id=subscription_id,
                product_block_id=self.product_block_id,
            )
            db.session.add(subscription_instance)
            db.session.flush()
            subscription_instances[self.subscription_instance_id] = subscription_instance

        if subscription_instance is None:
            raise ValueError("A subscription_id or subscription_instance argument is required")

        assert subscription_id

        # Everything is ok, make sure we are of the right class
        specialized_type = lookup_specialized_type(self.__class__, status)
        if specialized_type and not isinstance(self, specialized_type):
            raise ValueError(
                f"Lifecycle status {status} requires specialized type {specialized_type!r}, was: {type(self)!r}"
            )

        # Actually save stuff
        subscription_instance.label = self.label
        subscription_instance.values = self._save_instance_values(
            subscription_instance.product_block, subscription_instance.values
        )

        sub_instances = self._save_instances(subscription_id, subscription_instances, status)

        children: List[SubscriptionInstanceTable] = []
        for _domain_model_attr, instances in sub_instances.items():
            children.extend(instances)  # type: ignore
        subscription_instance.children = children

        # Make sure the db creates the relations
        db.session.flush()

        # Save the domain model attributes.
        self._set_instance_domain_model_attrs(sub_instances)

        return children + [subscription_instance]


class ProductModel(BaseModel):
    """Represent the product as defined in the database as a dataclass."""

    class Config:
        validate_assignment = True
        validate_all = True
        arbitrary_types_allowed = True

    product_id: UUID
    name: str
    description: str
    product_type: str
    tag: str
    status: ProductLifecycle


class SubscriptionModel(DomainModel):
    """Base class for all product subscription models.

    Define a subscription model:
    >>> class SubscriptionInactive(SubscriptionModel, product_type="SP"):  # doctest:+SKIP
    ...    block: Optional[ProductBlockModelInactive] = None

    >>> class Subscription(BlockInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):  # doctest:+SKIP
    ...    block: ProductBlockModel


    This example defines a subscription model with two different contraints based on lifecycle. `Subscription` is valid only for `ACTIVE`
    And `SubscriptionInactive` for all other states.
    `product_type` must be defined on the base class and need not to be defined on the others

    Create a new empty subscription
    >>> example1 = SubscriptionInactive.from_product_id(product_id, customer_id)  # doctest:+SKIP

    Create a new instance based on a dict in the state:
    >>> example2 = SubscriptionInactive(**state)  # doctest:+SKIP

    To retrieve a ProductBlockModel from the database:
    >>> SubscriptionInactive.from_subscription(subscription_id)  # doctest:+SKIP
    """

    product: ProductModel
    customer_id: UUID
    subscription_id: UUID = Field(default_factory=uuid4)
    description: str = "Initial subscription"
    status: SubscriptionLifecycle = SubscriptionLifecycle.INITIAL
    insync: bool = False
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    note: Optional[str] = None

    def __new__(cls, *args: Any, status: Optional[SubscriptionLifecycle] = None, **kwargs: Any) -> "SubscriptionModel":

        # status can be none if created during change_lifecycle
        if status and not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for status {status}")

        return super().__new__(cls)

    def __init_subclass__(
        cls, is_base: bool = False, lifecycle: Optional[List[SubscriptionLifecycle]] = None, **kwargs: Any
    ) -> None:
        super().__init_subclass__(lifecycle=lifecycle, **kwargs)

        if is_base:
            cls.__base_type__ = cls

        if is_base or lifecycle:
            register_specialized_type(cls, lifecycle)

        cls.__doc__ = make_subscription_model_docstring(cls, lifecycle)

    @classmethod
    def diff_product_in_database(cls, product_id: UUID) -> Dict[str, Any]:
        """Return any differences between the attrs defined on the domain model and those on product blocks in the database.

        This is only needed to check if the domain model and database models match which would be done during testing...
        """
        product_db = ProductTable.query.get(product_id)
        fixed_inputs_db = {fi.name for fi in product_db.fixed_inputs}
        fixed_inputs_model = set(cls._non_product_block_fields_.keys())
        missing_fixed_inputs_in_db = fixed_inputs_model - fixed_inputs_db
        missing_fixed_inputs_in_model = fixed_inputs_db - fixed_inputs_model
        logger.debug(
            "Fixed inputs diff",
            fixed_inputs_db=fixed_inputs_db,
            fixed_inputs_model=fixed_inputs_model,
            missing_fixed_inputs_in_db=missing_fixed_inputs_in_db,
            missing_fixed_inputs_in_model=missing_fixed_inputs_in_model,
        )

        def find_product_block_in(cls: Type[DomainModel]) -> List[ProductBlockModel]:
            product_blocks_in_model = []
            for product_block_field_type in cls._product_block_fields_.values():
                if is_list_type(product_block_field_type) or is_optional_type(product_block_field_type):
                    product_block_model = product_block_field_type.__args__[0]
                else:
                    product_block_model = product_block_field_type

                if issubclass(product_block_model, ProductBlockModel):
                    product_blocks_in_model.extend(find_product_block_in(product_block_model))

                product_blocks_in_model.append(product_block_model)
            return product_blocks_in_model

        product_blocks_in_db: Dict[str, ProductBlockTable] = {pb.name: pb for pb in product_db.product_blocks}
        product_blocks_in_model: List[ProductBlockModel] = find_product_block_in(cls)
        missing_product_blocks_in_db: List[str] = []

        missing_resource_types_in_db: Dict[str, Set[str]] = {}
        missing_resource_types_in_model: Dict[str, Set[str]] = {}

        # Check resource types and missing product blocks in model
        for product_block_model in product_blocks_in_model:
            resource_types_model = set(product_block_model._non_product_block_fields_)

            if product_block_model.name not in product_blocks_in_db:
                missing_product_blocks_in_db.append(product_block_model.name)
                missing_resource_types_in_db[product_block_model.name] = resource_types_model
                continue

            matching_product_block = product_blocks_in_db[product_block_model.name]

            logger.debug(
                "resource types for in model.",
                product_block_name=product_block_model.name,
                resource_types_model=resource_types_model,
            )

            logger.debug("Matching product block in database", product_block_name=matching_product_block.name)

            resource_types_db = {rt.resource_type for rt in matching_product_block.resource_types}
            logger.debug("Resource types in database.", resource_types_db=resource_types_db)
            missing_db = resource_types_model - resource_types_db
            if missing_db:
                missing_resource_types_in_db[matching_product_block.name] = missing_db

            missing_model = resource_types_db - resource_types_model
            if missing_model:
                missing_resource_types_in_model[matching_product_block.name] = missing_model

        missing_product_blocks_in_model = set(product_blocks_in_db.keys()).difference(
            map(attrgetter("name"), product_blocks_in_model)
        )

        logger.debug(
            "ProductTable blocks diff",
            missing_product_blocks_in_db=missing_product_blocks_in_db,
            missing_product_blocks_in_model=missing_product_blocks_in_model,
        )

        return {
            k: v
            for k, v in (
                ("missing_fixed_inputs_in_db", missing_fixed_inputs_in_db),
                ("missing_fixed_inputs_in_model", missing_fixed_inputs_in_model),
                ("missing_product_blocks_in_db", missing_product_blocks_in_db),
                ("missing_product_blocks_in_model", missing_product_blocks_in_model),
                ("missing_resource_types_in_db", missing_resource_types_in_db),
                ("missing_resource_types_in_model", missing_resource_types_in_model),
            )
            if v
        }

    @classmethod
    def from_product_id(
        cls: Type[S],
        product_id: Union[UUID, UUIDstr],
        customer_id: Union[UUID, UUIDstr],
        status: SubscriptionLifecycle = SubscriptionLifecycle.INITIAL,
        description: Optional[str] = None,
        insync: bool = False,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        note: Optional[str] = None,
    ) -> S:
        """Use product_id (and customer_id) to return required fields of a new empty subscription."""
        # Caller wants a new instance and provided a product_id and customer_id
        product_db = ProductTable.query.get(product_id)
        product = ProductModel(
            product_id=product_db.product_id,
            name=product_db.name,
            description=product_db.description,
            product_type=product_db.product_type,
            tag=product_db.tag,
            status=product_db.status,
        )

        if description is None:
            description = f"Initial subscription of {product.description}"

        fixed_inputs = {fi.name: fi.value for fi in product_db.fixed_inputs}
        instances = cls._init_instances()

        return cls(
            product=product,
            customer_id=customer_id,  # type:ignore
            description=description,
            status=status,
            insync=insync,
            start_date=start_date,
            end_date=end_date,
            note=note,
            **fixed_inputs,
            **instances,
        )

    @classmethod
    def from_subscription(cls: Type[S], subscription_id: Union[UUID, UUIDstr]) -> S:
        """Use a subscription_id to return required fields of an existing subscription."""
        subscription = SubscriptionTable.query.options(
            selectinload(SubscriptionTable.instances)
            .selectinload(SubscriptionInstanceTable.product_block)
            .selectinload(ProductBlockTable.resource_types),
            selectinload(SubscriptionTable.instances).selectinload(SubscriptionInstanceTable.parent_relations),
            selectinload(SubscriptionTable.instances).selectinload(SubscriptionInstanceTable.values),
        ).get(subscription_id)
        product = ProductModel(
            product_id=subscription.product.product_id,
            name=subscription.product.name,
            description=subscription.product.description,
            product_type=subscription.product.product_type,
            tag=subscription.product.tag,
            status=subscription.product.status,
        )
        status = SubscriptionLifecycle(subscription.status)

        if not hasattr(cls, "__base_type__"):
            old_cls = cls

            # Import here to prevent cyclic imports
            from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

            cls = SUBSCRIPTION_MODEL_REGISTRY.get(subscription.product.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)
            if cls != old_cls and not issubclass(cls, old_cls):
                raise ValueError(
                    f"{cls} is not a valid subclass of {old_cls}, make sure the class matches with the status in database"
                )

        fixed_inputs = {fi.name: fi.value for fi in subscription.product.fixed_inputs}
        instances = cls._load_instances(subscription.instances, status)

        try:
            return cls(  # type: ignore
                product=product,
                customer_id=subscription.customer_id,
                subscription_id=subscription_id,  # type:ignore
                description=subscription.description,
                status=status,
                insync=subscription.insync,
                start_date=subscription.start_date,
                end_date=subscription.end_date,
                note=subscription.note,
                **fixed_inputs,
                **instances,
            )
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database", loaded_fixed_inputs=fixed_inputs, loaded_instances=instances
            )
            raise

    def save(self) -> None:
        """Save the subscription to the database."""
        specialized_type = lookup_specialized_type(self.__class__, self.status)
        if specialized_type and not isinstance(self, specialized_type):
            raise ValueError(
                f"Lifecycle status {self.status.value} requires specialized type {specialized_type!r}, was: {type(self)!r}"
            )

        sub = SubscriptionTable.query.options(
            selectinload(SubscriptionTable.instances)
            .selectinload(SubscriptionInstanceTable.product_block)
            .selectinload(ProductBlockTable.resource_types),
            selectinload(SubscriptionTable.instances).selectinload(SubscriptionInstanceTable.values),
        ).get(self.subscription_id)
        if sub:
            sub.product_id = self.product.product_id
            sub.customer_id = self.customer_id
            sub.description = self.description
            sub.status = self.status.value
            sub.insync = self.insync
            sub.start_date = self.start_date
            sub.end_date = self.end_date
            sub.note = self.note
            db.session.add(sub)
        else:
            sub = SubscriptionTable(
                subscription_id=self.subscription_id,
                product_id=self.product.product_id,
                customer_id=self.customer_id,
                description=self.description,
                status=self.status.value,
                insync=self.insync,
                start_date=self.start_date,
                end_date=self.end_date,
                note=self.note,
            )

            db.session.add(sub)
            db.session.flush()  # Sends INSERT and returns subscription_id without committing transaction

        old_instances_dict = {instance.subscription_instance_id: instance for instance in sub.instances}

        new_instance_dict = self._save_instances(self.subscription_id, old_instances_dict, sub.status)

        saved_instances: List[SubscriptionInstanceTable] = []
        for _key, values in new_instance_dict.items():
            saved_instances.extend(values)  # type: ignore

        sub.instances = saved_instances

        # Calculate what to remove
        instances_set = {instance.subscription_instance_id for instance in sub.instances}
        for instance_id in instances_set:
            old_instances_dict.pop(instance_id, None)

        # What's left should be removed
        for instance in old_instances_dict.values():
            db.session.delete(instance)

        db.session.flush()


SI = TypeVar("SI")


class SubscriptionInstanceList(ConstrainedList, List[SI]):
    """Shorthand to create constrained lists of product blocks."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)  # type:ignore

        # Copy generic argument (SI) if not set explicitly
        # This makes a lot of assuptions about the internals of `typing`
        if "__orig_bases__" in cls.__dict__ and cls.__dict__["__orig_bases__"]:
            generic_base_cls = cls.__dict__["__orig_bases__"][0]
            if (
                not hasattr(generic_base_cls, "item_type")
                and hasattr(generic_base_cls, "__args__")
                and generic_base_cls.__args__
            ):
                cls.item_type = generic_base_cls.__args__[0]

        # Make sure __args__ is set
        cls.__args__ = (cls.item_type,)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple, Type, TypeVar

import structlog

from orchestrator.types import SubscriptionLifecycle, strEnum
from orchestrator.utils.datetime import nowtz

logger = structlog.get_logger(__name__)


class ProductLifecycle(strEnum):
    ACTIVE = "active"
    PRE_PRODUCTION = "pre production"
    PHASE_OUT = "phase out"
    END_OF_LIFE = "end of life"


_sub_type_per_lifecycle: Dict[Tuple[Type, Optional[SubscriptionLifecycle]], Type] = {}


def register_specialized_type(cls: Type, lifecycle: Optional[List[SubscriptionLifecycle]] = None) -> None:
    if lifecycle:
        for lifecycle_state in lifecycle:
            _sub_type_per_lifecycle[(cls.__base_type__, lifecycle_state)] = cls
    else:
        _sub_type_per_lifecycle[(cls.__base_type__, None)] = cls


def lookup_specialized_type(block: Type, lifecycle: Optional[SubscriptionLifecycle]) -> Type:
    if not hasattr(block, "__base_type__"):
        raise ValueError("Cannot instantiate a class that has no __base_type__ attribute")

    specialized_block = _sub_type_per_lifecycle.get((block.__base_type__, lifecycle), None)
    if specialized_block is None:
        specialized_block = _sub_type_per_lifecycle.get((block.__base_type__, None), None)
    if specialized_block is None:
        specialized_block = block
    return specialized_block


T = TypeVar("T")


def change_lifecycle(subscription: T, status: SubscriptionLifecycle) -> T:
    new_klass = lookup_specialized_type(subscription.__class__, status)
    data = subscription.dict()  # type:ignore

    data["status"] = status
    if data["start_date"] is None and status == SubscriptionLifecycle.ACTIVE:
        data["start_date"] = nowtz()
    if data["end_date"] is None and status == SubscriptionLifecycle.TERMINATED:
        data["end_date"] = nowtz()

    return new_klass(**data)



# type: ignore

from typing import List, Optional
from uuid import uuid4

import pytest

from orchestrator.domain.lifecycle import change_lifecycle
from orchestrator.types import State, SubscriptionLifecycle
from orchestrator.utils.state import extract, inject_args

STATE = {"one": 1, "two": 2, "three": 3, "four": 4}


def test_extract():
    one, two, three, four = extract(("one", "two", "three", "four"), STATE)
    assert one == 1
    assert two == 2
    assert three == 3
    assert four == 4

    four, three, two, one = extract(("four", "three", "two", "one"), STATE)
    assert one == 1
    assert two == 2
    assert three == 3
    assert four == 4

    nothing = extract((), STATE)
    assert len(nothing) == 0


def test_extract_key_error():
    key = "I don't exist"
    with pytest.raises(KeyError) as excinfo:
        extract((key,), STATE)
        assert key in excinfo.value.args


def test_state() -> None:
    @inject_args
    def step_func_ok(one):
        assert one == STATE["one"]
        return {"prefix_id": 42}

    new_state = step_func_ok(STATE)
    assert "prefix_id" in new_state
    assert new_state["prefix_id"] == 42

    @inject_args
    def step_func_fail(i_am_not_in_the_state):
        return {}

    with pytest.raises(KeyError):
        step_func_fail(STATE)

    @inject_args
    def step_func_opt_arg(opt: Optional[str] = None) -> None:
        assert opt is None

    step_func_opt_arg(STATE)

    @inject_args
    def step_func_default(default="bla"):
        assert default == "bla"

    step_func_default(STATE)


def test_inject_args(generic_product_1, generic_product_type_1) -> None:
    GenericProductOneInactive, GenericProduct = generic_product_type_1
    product_id = generic_product_1.product_id
    state = {"product": product_id, "organisation": uuid4()}
    generic_sub = GenericProductOneInactive.from_product_id(
        product_id=state["product"], customer_id=state["organisation"], status=SubscriptionLifecycle.INITIAL
    )
    generic_sub.pb_1.rt_1 = "test"
    generic_sub.pb_2.rt_2 = 42
    generic_sub.pb_2.rt_3 = "test2"

    generic_sub = change_lifecycle(generic_sub, SubscriptionLifecycle.ACTIVE)

    generic_sub.save()

    @inject_args
    def step_existing(generic_sub: GenericProduct) -> State:
        assert generic_sub.subscription_id
        assert generic_sub.pb_1.rt_1 == "test"
        generic_sub.pb_1.rt_1 = "test string"
        return {"generic_sub": generic_sub}

    # Put `generic_sub` as an UUID in. Entire `generic_sub` object would have worked as well, but this way we will be
    # certain that if we end up with an entire `generic_sub` object in the step function, it will have been retrieved
    # from the database.
    state["generic_sub"] = generic_sub.subscription_id

    state_amended = step_existing(state)
    assert "generic_sub" in state_amended

    # Do we now have an entire object instead of merely a UUID
    assert isinstance(state_amended["generic_sub"], GenericProduct)

    # And does it have the modifcations from the step functions
    assert state_amended["generic_sub"].pb_1.rt_1 == "test string"

    # Test `rt_1` has been persisted to the database with the modifications from the step function.`
    fresh_generic_sub = GenericProduct.from_subscription(state_amended["generic_sub"].subscription_id)
    assert fresh_generic_sub.pb_1.rt_1 is not None


def test_inject_args_list(generic_product_1, generic_product_type_1) -> None:
    GenericProductOneInactive, GenericProduct = generic_product_type_1
    product_id = generic_product_1.product_id
    state = {"product": product_id, "organisation": uuid4()}
    generic_sub = GenericProductOneInactive.from_product_id(
        product_id=state["product"], customer_id=state["organisation"], status=SubscriptionLifecycle.INITIAL
    )
    generic_sub.pb_1.rt_1 = "test"
    generic_sub.pb_2.rt_2 = 42
    generic_sub.pb_2.rt_3 = "test2"

    generic_sub = change_lifecycle(generic_sub, SubscriptionLifecycle.ACTIVE)

    generic_sub.save()

    @inject_args
    def step_existing(generic_sub: List[GenericProduct]) -> State:
        assert len(generic_sub) == 1
        assert generic_sub[0].subscription_id
        assert generic_sub[0].pb_1.rt_1 == "test"
        return {"generic_sub": generic_sub}

    # Put `generic_sub` as an UUID in. Entire `generic_sub` object would have worked as well, but this way we will be
    # certain that if we end up with an entire `generic_sub` object in the step function, it will have been retrieved
    # from the database.
    state["generic_sub"] = [generic_sub.subscription_id]

    state_amended = step_existing(state)
    assert "generic_sub" in state_amended
    assert len(state_amended["generic_sub"]) == 1

    # Do we now have an entire object instead of merely a UUID
    assert isinstance(state_amended["generic_sub"][0], GenericProduct)

    # And does it have the modifcations from the step functions
    assert state_amended["generic_sub"][0].pb_1.rt_1 is not None

    # Test `rt_1` has been persisted to the database with the modifications from the step function.`
    fresh_generic_sub = GenericProduct.from_subscription(state_amended["generic_sub"][0].subscription_id)
    assert fresh_generic_sub.pb_1.rt_1 is not None


def test_inject_args_optional(generic_product_1, generic_product_type_1) -> None:
    GenericProductOneInactive, GenericProduct = generic_product_type_1
    product_id = generic_product_1.product_id
    state = {"product": product_id, "organisation": uuid4()}
    generic_sub = GenericProductOneInactive.from_product_id(
        product_id=state["product"], customer_id=state["organisation"], status=SubscriptionLifecycle.INITIAL
    )
    generic_sub.pb_1.rt_1 = "test"
    generic_sub.pb_2.rt_2 = 42
    generic_sub.pb_2.rt_3 = "test2"

    generic_sub = change_lifecycle(generic_sub, SubscriptionLifecycle.ACTIVE)

    generic_sub.save()

    @inject_args
    def step_existing(generic_sub: Optional[GenericProduct]) -> State:
        assert generic_sub is not None, "Generic Sub IS NONE"
        assert generic_sub.subscription_id
        assert generic_sub.pb_1.rt_1 == "test"
        return {"generic_sub": generic_sub}

    with pytest.raises(AssertionError) as exc_info:
        step_existing(state)

    assert "Generic Sub IS NONE" in str(exc_info.value)

    # Put `light_path` as an UUID in. Entire `light_path` object would have worked as well, but this way we will be
    # certain that if we end up with an entire `light_path` object in the step function, it will have been retrieved
    # from the database.
    state["generic_sub"] = generic_sub.subscription_id

    state_amended = step_existing(state)
    assert "generic_sub" in state_amended

    # Do we now have an entire object instead of merely a UUID
    assert isinstance(state_amended["generic_sub"], GenericProduct)

    # And does it have the modifcations from the step functions
    assert state_amended["generic_sub"].pb_1.rt_1 is not None

    # Test `nso_service_id` has been persisted to the database with the modifications from the step function.`
    fresh_generic_sub = GenericProduct.from_subscription(state_amended["generic_sub"].subscription_id)
    assert fresh_generic_sub.pb_1.rt_1 is not None



from http import HTTPStatus
from uuid import uuid4

import pytest

from orchestrator.config import CITY_TYPE, DOMAIN, IMS_CIRCUIT_ID, PORT_SPEED, PORT_SUBSCRIPTION_ID, SERVICE_SPEED
from orchestrator.db import (
    FixedInputTable,
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.utils.json import json_dumps

PRODUCT_ID = uuid4()
MSP_PRODUCT_ID = uuid4()
UNPROTECTED_MSP_SSP_ID = uuid4()
PROTECTED_MSP_SSP_ID = uuid4()
REDUNDANT_MSP_SSP_ID = uuid4()
SUBSCRIPTION_ID = uuid4()
CHILD_SUBSCRIPTION_ID = uuid4()


@pytest.fixture
def seed():
    ProductTable.query.delete()
    resources = [
        ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="IMS circuit Id"),
    ]
    product_blocks = [ProductBlockTable(name="Ethernet", description="d", status="active", resource_types=resources)]
    fixed_inputs = [
        FixedInputTable(name=SERVICE_SPEED, value="1000"),
        FixedInputTable(name=CITY_TYPE, value="inner_city"),
        FixedInputTable(name="protection_type", value="Protected"),
    ]
    product = ProductTable(
        product_id=PRODUCT_ID,
        name="LightPath",
        description="descr",
        product_type="LightPath",
        tag="LightPath",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    msp_product = ProductTable(
        product_id=MSP_PRODUCT_ID,
        name="MSP",
        description="descr",
        product_type="Port",
        tag="Port",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=[FixedInputTable(name=PORT_SPEED, value="1000"), FixedInputTable(name=DOMAIN, value="SURFNET7")],
    )
    msp_product_100G = ProductTable(
        name="MSP 1100G",
        description="descr",
        product_type="Port",
        tag="Port",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=[
            FixedInputTable(name=PORT_SPEED, value="100000"),
            FixedInputTable(name="domain", value="SURFNET7"),
        ],
    )
    product_no_workflow = ProductTable(
        product_id=uuid4(), name="P", description="P", product_type="Port", tag="P", status="active"
    )

    port_subscription_id = ResourceTypeTable(resource_type=PORT_SUBSCRIPTION_ID, description="Port Subscription Id")
    values = [SubscriptionInstanceValueTable(resource_type=port_subscription_id, value=str(CHILD_SUBSCRIPTION_ID))]
    subscription = SubscriptionTable(
        subscription_id=SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=product,
        customer_id=uuid4(),
        instances=[SubscriptionInstanceTable(product_block=product_blocks[0], values=values)],
    )
    child_subscription = SubscriptionTable(
        subscription_id=CHILD_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=msp_product,
        customer_id=uuid4(),
    )

    def lp_product(protection_type, product_id, speed="1000"):
        name = f"msp_ssp_{protection_type.lower()}"
        msp_ssp_fixed_inputs = [
            FixedInputTable(name="protection_type", value=protection_type.capitalize()),
            FixedInputTable(name=SERVICE_SPEED, value=speed),
            FixedInputTable(name=CITY_TYPE, value="inner_city"),
        ]
        return ProductTable(
            product_id=product_id,
            name=f"name_{product_id}",
            description=name,
            product_type="LightPath",
            tag="LightPath",
            status="active",
            fixed_inputs=msp_ssp_fixed_inputs,
        )

    db.session.add(product)
    db.session.add(msp_product)
    db.session.add(msp_product_100G)
    db.session.add(lp_product("protected", uuid4(), speed="10000"))
    db.session.add(lp_product("protected", uuid4(), speed="750"))
    db.session.add(lp_product("redundant", REDUNDANT_MSP_SSP_ID))
    db.session.add(lp_product("protected", PROTECTED_MSP_SSP_ID, speed="250"))
    db.session.add(lp_product("unprotected", UNPROTECTED_MSP_SSP_ID))
    db.session.add(subscription)
    db.session.add(child_subscription)
    db.session.add(product_no_workflow)
    db.session.commit()


def test_fetch_all_products(seed, test_client):
    response = test_client.get("/api/products")

    assert HTTPStatus.OK == response.status_code
    products = response.json()

    assert 9 == len(products)


def test_fetch_all_products_by_tag_and_port(seed, test_client):
    response = test_client.get("/api/products?tag=LightPath&product_type=LightPath")

    assert HTTPStatus.OK == response.status_code
    products = response.json()

    assert 6 == len(products)


def test_fetch_all_products_by_unknown_tag(seed, test_client):
    response = test_client.get("/api/products?tag=Nope")

    assert HTTPStatus.OK == response.status_code
    products = response.json()

    assert 0 == len(products)


def test_product_by_id(seed, test_client):
    response = test_client.get(f"/api/products/{PRODUCT_ID}")

    assert HTTPStatus.OK == response.status_code
    product = response.json()
    expected_value = [fi for fi in product["fixed_inputs"] if fi["name"] == SERVICE_SPEED][0]["value"]
    assert "1000" == expected_value
    assert "Ethernet" == product["product_blocks"][0]["name"]


def test_product_by_id_404(seed, test_client):
    response = test_client.get(f"/api/products/{str(uuid4())}")
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_save(seed, test_client):
    p_id = uuid4()
    body = {
        "product_id": p_id,
        "name": "MSP_2",
        "description": "MSP",
        "product_type": "Port",
        "tag": "Port",
        "status": "active",
        "fixed_inputs": [{"name": "help", "value": "val"}],
        "product_blocks": [
            {"name": "name", "description": "desc", "status": "active", "resource_types": [{"resource_type": "test"}]}
        ],
    }
    from orchestrator.db import FixedInputTable

    fi = FixedInputTable.query.all()
    response = test_client.post("/api/products/", data=json_dumps(body), headers={"Content_Type": "application/json"})
    assert HTTPStatus.NO_CONTENT == response.status_code
    products = test_client.get("/api/products").json()
    assert 10 == len(products)

    fi2 = FixedInputTable.query.all()
    assert len(fi) + 1 == len(fi2)


def test_delete(seed, test_client):
    p_id = uuid4()
    body = {
        "product_id": p_id,
        "name": "DEL",
        "description": "DEL",
        "product_type": "Port",
        "tag": "Port",
        "status": "active",
    }
    test_client.post("/api/products/", data=json_dumps(body), headers={"Content_Type": "application/json"})

    response = test_client.delete(f"/api/products/{p_id}")
    assert HTTPStatus.NO_CONTENT == response.status_code


def test_delete_with_subscriptions(seed, test_client):
    response = test_client.delete(f"/api/products/{PRODUCT_ID}")

    assert HTTPStatus.BAD_REQUEST == response.status_code
    assert f"Product {PRODUCT_ID} is used in Subscriptions: LightPath" == response.json()["detail"]


def test_tags_all(test_client):
    response = test_client.get("/api/products/tags/all")

    assert HTTPStatus.OK == response.status_code
    tags = response.json()

    tags_db = [p.tag for p in ProductTable.query.distinct(ProductTable.tag).all()]
    assert set(tags) == set(tags_db)


def test_types_all(test_client):
    response = test_client.get("/api/products/types/all")

    assert HTTPStatus.OK == response.status_code
    types = response.json()
    types_db = [p.product_type for p in ProductTable.query.distinct(ProductTable.product_type).all()]
    assert set(types) == set(types_db)


def test_statuses_all(seed, test_client):
    response = test_client.get("/api/products/statuses/all")

    assert HTTPStatus.OK == response.status_code
    statuses = response.json()
    assert {"active", "phase out", "pre production", "end of life"}.issubset(set(statuses))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Fixed inputs
CITY_TYPE = "city_type"
DOMAIN = "domain"
PORT_SPEED = "port_speed"
PROTECTION_TYPE = "protection_type"
SERVICE_SPEED = "service_speed"
AGGREGATE = "aggregate"
REDUNDANT = "redundant"
TAGGED = "tagged"
NETHERLIGHT = "netherlight"
IP_ROUTING_TYPE = "ip_routing_type"
SIZE = "size"

# Resource types
PORT_SUBSCRIPTION_ID = "port_subscription_id"
IMS_CIRCUIT_ID = "ims_circuit_id"
IMS_NODE_ID = "ims_node_id"
IPAM_PREFIX_ID = "ipam_prefix_id"
IP_PREFIX_SUBSCRIPTION_ID = "ip_prefix_subscription_id"
PARENT_IP_PREFIX_SUBSCRIPTION_ID = "parent_ip_prefix_subscription_id"
NSO_DEVICE_ID = "nso_device_id"  # will contain the NSO device name
NSO_SERVICE_ID = "nso_service_id"  # will contain the service UUID
NODE_SUBSCRIPTION_ID = "node_subscription_id"
PORT_MODE = "port_mode"
INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID = "internetpinnen_prefix_subscription_id"
PEER_GROUP_SUBSCRIPTION_ID = "peer_group_subscription_id"

MAX_SPEED_POSSIBLE = 1000_000



from orchestrator.config.assignee import Assignee
from orchestrator.forms import FormPage
from orchestrator.workflow import begin, done, inputstep, step, workflow
from test.unit_tests.workflows import (
    WorkflowInstanceForTests,
    assert_complete,
    assert_suspended,
    resume_workflow,
    run_workflow,
)


def test_resume_workflow():
    @step("Test step 1")
    def fakestep1():
        return {"steden": ["Utrecht"], "stad": "Amsterdam"}

    @inputstep("Wait for me step 2", assignee=Assignee.SYSTEM)
    def waitforme(steden, stad):
        class WaitForm(FormPage):
            stad: str

        user_input = yield WaitForm
        meer = [*steden, stad]
        return {**user_input.dict(), "steden": meer}

    @step("Test step 3")
    def fakestep3(steden):
        meer = [*steden, "Leiden"]
        return {"steden": meer}

    @workflow("Test wf", target="Target.CREATE")
    def testwf():
        return begin >> fakestep1 >> waitforme >> fakestep3 >> done

    WorkflowInstanceForTests(testwf, "testwf")

    # The process and step_log from run_workflow can be used to resume it with resume_workflow
    result, process, step_log = run_workflow("testwf", {})
    assert_suspended(result)
    result, step_log = resume_workflow(process, step_log, {"stad": "Maastricht"})
    assert_complete(result)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements subscription related API endpoints."""

from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

import structlog
from fastapi.routing import APIRouter
from sqlalchemy import Text, and_, cast, or_, text
from sqlalchemy.orm import aliased, contains_eager, defer, joinedload, undefer
from sqlalchemy.orm.exc import MultipleResultsFound
from starlette.responses import Response

from orchestrator.api.error_handling import raise_status
from orchestrator.api.helpers import _query_with_filters
from orchestrator.db import (
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProcessTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionCustomerDescriptionTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.domain.base import SubscriptionModel
from orchestrator.schemas import (
    SubscriptionDetailsSchema,
    SubscriptionDomainModelSchema,
    SubscriptionIdSchema,
    SubscriptionSchema,
    SubscriptionWithPortAttrSchema,
    SubscriptionWorkflowListsSchema,
)
from orchestrator.services.subscriptions import (
    RELATION_RESOURCE_TYPES,
    get_subscription_details_by_id,
    query_child_subscriptions_by_resource_types,
    query_parent_subscriptions_by_resource_types,
    subscription_workflows,
)

router = APIRouter()

logger = structlog.get_logger(__name__)


def _delete_subscription_tree(subscription: SubscriptionTable) -> None:
    db.session.delete(subscription)
    db.session.commit()


def _delete_process_subscriptions(process_subscriptions: List[ProcessSubscriptionTable]) -> None:
    for process_subscription in process_subscriptions:
        pid = str(process_subscription.pid)
        subscription_id = str(process_subscription.subscription_id)
        ProcessSubscriptionTable.query.filter(ProcessSubscriptionTable.pid == pid).delete()
        ProcessStepTable.query.filter(ProcessStepTable.pid == pid).delete()
        ProcessTable.query.filter(ProcessTable.pid == pid).delete()
        subscription = SubscriptionTable.query.filter(SubscriptionTable.subscription_id == subscription_id).first()
        _delete_subscription_tree(subscription)


@router.get("/product_type/{product_type}", response_model=List[SubscriptionWithPortAttrSchema])
def subscriptions_by_product_type(product_type: str) -> List[SubscriptionWithPortAttrSchema]:
    subs = (
        SubscriptionTable.query.join(ProductTable)
        .options(undefer(SubscriptionTable.name), undefer(SubscriptionTable.tag), undefer(SubscriptionTable.port_mode))
        .filter(ProductTable.product_type == product_type)
        .all()
    )
    return subs


@router.get("/product_type/{product_type}/customer/{customer_id}", response_model=List[SubscriptionSchema])
def by_product_type_and_customer(product_type: str, customer_id: UUID) -> List[SubscriptionSchema]:
    """
    Retrieve Subscription by product type and customer id.

    This EndpointSchema is for the new Networkdashboard.

    :param product_type: Product Type out of the database.
    :param customer_id: Customer that is being queried.
    :return: List of Subscriptions
    """

    if product_type in ["LightPath", "IP", "L2VPN", "Firewall"]:
        # First retrieve all LightPaths, IP and L2VPN subscriptions

        product_type_filter = (
            or_(ProductTable.product_type == "IP", ProductTable.product_type == "Firewall")
            if product_type == "IP"
            else ProductTable.product_type == product_type
        )
        logger.debug("ProductTable type filter", product_type_filter=product_type_filter)
        subs = (
            SubscriptionTable.query.join(ProductTable)
            .options(
                undefer(SubscriptionTable.name),
                undefer(SubscriptionTable.tag),
                undefer(SubscriptionTable.port_mode),
                joinedload(SubscriptionTable.product),
                joinedload(SubscriptionTable.customer_descriptions),
            )
            .filter(SubscriptionTable.customer_id == customer_id)
            .filter(product_type_filter)
            .all()
        )

        port_subs = (
            db.session.query(cast(SubscriptionTable.subscription_id, Text))
            .join(ProductTable)
            .filter(SubscriptionTable.customer_id == customer_id, ProductTable.product_type == "Port")
            .subquery()
        )

        service_subs = (
            SubscriptionTable.query.join(ProductTable, SubscriptionInstanceTable, SubscriptionInstanceValueTable)
            .options(
                undefer(SubscriptionTable.name),
                undefer(SubscriptionTable.tag),
                undefer(SubscriptionTable.port_mode),
                joinedload(SubscriptionTable.product),
                joinedload(SubscriptionTable.customer_descriptions),
            )
            .filter(product_type_filter, SubscriptionInstanceValueTable.value.in_(port_subs))
            .all()
        )
        subs += service_subs

        return list(set(subs))
    else:
        subs = (
            SubscriptionTable.query.join(ProductTable)
            .options(
                undefer(SubscriptionTable.name),
                undefer(SubscriptionTable.tag),
                undefer(SubscriptionTable.port_mode),
                joinedload(SubscriptionTable.product),
                joinedload(SubscriptionTable.customer_descriptions),
            )
            .filter(SubscriptionTable.customer_id == customer_id)
            .filter(ProductTable.product_type == product_type)
            .all()
        )

        return list(set(subs))


@router.get("/customer/{customer_id}", response_model=List[SubscriptionWithPortAttrSchema])
def subscriptions_by_customer(customer_id: UUID) -> List[SubscriptionWithPortAttrSchema]:
    subs = (
        SubscriptionTable.query.options(
            undefer(SubscriptionTable.name),
            undefer(SubscriptionTable.tag),
            undefer(SubscriptionTable.port_mode),
            joinedload(SubscriptionTable.product),
        )
        .filter(SubscriptionTable.customer_id == customer_id)
        .all()
    )
    other_services_terminated_at_customer_ports: List[SubscriptionTable] = []
    for sub in subs:
        if sub.product.product_type == "Port":
            port_subs = (
                SubscriptionTable.query.join(SubscriptionInstanceTable)
                .join(SubscriptionInstanceValueTable)
                .join(ResourceTypeTable)
                .options(
                    undefer(SubscriptionTable.name),
                    undefer(SubscriptionTable.tag),
                    undefer(SubscriptionTable.port_mode),
                    joinedload(SubscriptionTable.product),
                )
                .filter(ResourceTypeTable.resource_type == "port_subscription_id")
                .filter(SubscriptionInstanceValueTable.value == str(sub.subscription_id))
                .filter(SubscriptionTable.customer_id != customer_id)
                .all()
            )
            if port_subs:
                other_services_terminated_at_customer_ports.extend(port_subs)

    subs.extend(other_services_terminated_at_customer_ports)

    return subs


@router.get("/all", response_model=List[SubscriptionSchema])
def subscriptions_all() -> List[SubscriptionSchema]:
    """Return subscriptions with only a join on products."""
    subscription_columns = [
        "subscription_id",
        "description",
        "status",
        "insync",
        "start_date",
        "end_date",
        "customer_id",
    ]
    query = """SELECT s.subscription_id, s.description as description, s.status, s.insync, s.start_date,
               s.end_date, s.customer_id,
               p.created_at as product_created_at, p.description as product_description,
               p.end_date as product_end_date, p.name as product_name,
               p.tag as product_tag, p.product_id as product_product_id, p.status as product_status,
               p.product_type as product_type
                FROM subscriptions s
                 JOIN products p ON s.product_id = p.product_id"""
    subscriptions = db.session.execute(text(query))
    result: List[SubscriptionSchema] = []
    for sub in subscriptions:
        sub_dict = dict(zip(subscription_columns, sub))
        product = {
            "description": sub["product_description"],
            "name": sub["product_name"],
            "product_id": sub["product_product_id"],
            "status": sub["product_status"],
            "product_type": sub["product_type"],
            "tag": sub["product_tag"],
        }
        sub_dict["product"] = product
        result.append(SubscriptionSchema.parse_obj(sub_dict))
    # validation the response doesn't work correct for nested responses: a unit test ensures schema correctness
    return result


@router.get("/ports", response_model=List[SubscriptionWithPortAttrSchema])
def port_subscriptions_filterable(
    response: Response, range: Optional[str] = None, sort: Optional[str] = None, filter: Optional[str] = None
) -> List[SubscriptionTable]:
    _range: Union[List[int], None] = list(map(int, range.split(","))) if range else None
    _sort: Union[List[str], None] = sort.split(",") if sort else None
    _filter: Union[List[str], None] = filter.split(",") if filter else None
    logger.info("port_subscriptions_filterable() called", range=_range, sort=_sort, filter=_filter)
    query = (
        SubscriptionTable.query.join(SubscriptionTable.product)
        .options(contains_eager(SubscriptionTable.product), defer("product_id"), undefer("port_mode"))
        .filter(ProductTable.product_type == "Port")
    )
    query_result = _query_with_filters(response, query, _range, _sort, _filter)

    return query_result


@router.get("/nsi", response_model=SubscriptionIdSchema)
def nsi_light_path_subscription(
    port_subscription_id_1: UUID, vlan_1: int, port_subscription_id_2: UUID, vlan_2: int
) -> Dict[str, UUID]:
    """
    Retrieve an NSI LightPath SubscriptionTable ID for OpenNSA.

    This is a custom endpoint in the orchestrator to support the OpenNSA deployment. It is necessary to be compatible
    with OpenNSA. The OpenNSA only has a notion of interfaces and vlans, not of deployed LightPaths. It searches for
    deployed LightPaths with the port and vlan combination.

    Args:
        port_subscription_id_1: PortSchema one
        vlan_1: vlan 1
        port_subscription_id_2: PortSchema 2
        vlan_2: vlan 2

    Returns:
        {"subscription_id": UUID}

    """
    Port1InstanceValue = aliased(SubscriptionInstanceValueTable)
    Vlan1InstanceValue = aliased(SubscriptionInstanceValueTable)
    Port2InstanceValue = aliased(SubscriptionInstanceValueTable)
    Vlan2InstanceValue = aliased(SubscriptionInstanceValueTable)
    SAP1 = aliased(SubscriptionInstanceTable)
    SAP2 = aliased(SubscriptionInstanceTable)

    try:
        subscription = (
            SubscriptionTable.query.join(SAP1, SAP1.subscription_id == SubscriptionTable.subscription_id)
            .join(SAP2, SAP2.subscription_id == SubscriptionTable.subscription_id)
            .join(Port1InstanceValue, Port1InstanceValue.subscription_instance_id == SAP1.subscription_instance_id)
            .join(Vlan1InstanceValue, Vlan1InstanceValue.subscription_instance_id == SAP1.subscription_instance_id)
            .join(Port2InstanceValue, Port2InstanceValue.subscription_instance_id == SAP2.subscription_instance_id)
            .join(Vlan2InstanceValue, Vlan2InstanceValue.subscription_instance_id == SAP2.subscription_instance_id)
            .join(ProductTable)
            .filter(
                and_(
                    Port1InstanceValue.value == str(port_subscription_id_1),
                    Vlan1InstanceValue.value == str(vlan_1),
                    Port2InstanceValue.value == str(port_subscription_id_2),
                    Vlan2InstanceValue.value == str(vlan_2),
                    ProductTable.product_type == "LightPath",
                    or_(ProductTable.tag == "LP", ProductTable.tag == "LPNL"),
                    SubscriptionTable.status == "active",
                ),
            )
        ).one_or_none()
        if not subscription:
            raise_status(HTTPStatus.NOT_FOUND, "Combination not found")

    except MultipleResultsFound:
        logger.exception(
            "Multiple subscriptions found for query with following keys",
            port_subscription_id_1=port_subscription_id_1,
            vlan_1=vlan_1,
            port_subscription_id_2=port_subscription_id_2,
            vlan_2=vlan_2,
        )
        raise_status(HTTPStatus.INTERNAL_SERVER_ERROR, detail="Multiple records found")

    return {"subscription_id": subscription.subscription_id}


@router.get("/{subscription_id}", response_model=SubscriptionDetailsSchema)
def subscription_details_by_id(subscription_id: UUID) -> Optional[SubscriptionTable]:
    subscription = get_subscription_details_by_id(subscription_id)
    if not subscription:
        raise_status(HTTPStatus.NOT_FOUND)
    return subscription


@router.get("/domain-model/{subscription_id}", response_model=SubscriptionDomainModelSchema)
def subscription_details_by_id_with_domain_model(subscription_id: UUID) -> Dict[str, Any]:
    customer_descriptions = SubscriptionCustomerDescriptionTable.query.filter(
        SubscriptionCustomerDescriptionTable.subscription_id == subscription_id
    ).all()

    subscription = SubscriptionModel.from_subscription(subscription_id).dict()

    subscription["customer_descriptions"] = customer_descriptions

    if not subscription:
        raise_status(HTTPStatus.NOT_FOUND)
    return subscription


@router.delete("/{subscription_id}", response_model=None)
def delete_subscription(subscription_id: UUID) -> None:
    all_process_subscriptions = ProcessSubscriptionTable.query.filter_by(subscription_id=subscription_id).all()
    if len(all_process_subscriptions) > 0:
        _delete_process_subscriptions(all_process_subscriptions)
        return None
    else:
        subscription = SubscriptionTable.query.filter(SubscriptionTable.subscription_id == subscription_id).first()
        if not subscription:
            raise_status(HTTPStatus.NOT_FOUND)

        _delete_subscription_tree(subscription)
        return None


@router.get("/tag/{tags}", response_model=List[SubscriptionWithPortAttrSchema])
@router.get("/tag/{tags}/{statuses}", response_model=List[SubscriptionWithPortAttrSchema])
def subscriptions_by_tags(tags: str, statuses: Optional[str] = None) -> List[SubscriptionTable]:
    """Return subscriptions by product tags and optionally subscription statuses.

    Args:
        tags: Comma separated string of product tags
        statuses: comma separated string of subscriptions statuses

    Returns:
        List of subscriptions

    """
    query = (
        SubscriptionTable.query.join(ProductTable)
        .options(undefer(SubscriptionTable.name), undefer(SubscriptionTable.tag), undefer(SubscriptionTable.port_mode))
        .filter(ProductTable.tag.in_(list(map(lambda x: x.strip(), tags.split(",")))))
    )
    if statuses is not None:
        query = query.filter(SubscriptionTable.status.in_(list(map(lambda x: x.strip(), statuses.split(",")))))
    return query.all()


@router.get("/product/{product_id}", response_model=List[SubscriptionWithPortAttrSchema])
def subscriptions_by_product(product_id: UUID) -> List[SubscriptionTable]:
    return (
        SubscriptionTable.query.join(ProductTable)
        .options(undefer(SubscriptionTable.name), undefer(SubscriptionTable.tag), undefer(SubscriptionTable.port_mode))
        .filter(ProductTable.product_id == product_id)
        .filter(SubscriptionTable.status == "active")
        .all()
    )


@router.get("/parent_subscriptions/{subscription_id}", response_model=List[SubscriptionSchema])
def parent_subscriptions(subscription_id: UUID) -> List[SubscriptionTable]:
    return query_parent_subscriptions_by_resource_types(RELATION_RESOURCE_TYPES, subscription_id).all()


@router.get("/child_subscriptions/{subscription_id}", response_model=List[SubscriptionSchema])
def child_subscriptions(subscription_id: UUID) -> List[SubscriptionTable]:
    return query_child_subscriptions_by_resource_types(RELATION_RESOURCE_TYPES, subscription_id).all()


@router.get("/", response_model=List[SubscriptionSchema])
def subscriptions_filterable(
    response: Response, range: Optional[str] = None, sort: Optional[str] = None, filter: Optional[str] = None
) -> List[SubscriptionTable]:
    """
    Get subscriptions filtered.

    Args:
        response: Fastapi Response object
        range: Range
        sort: Sort
        filter: Filter

    Returns:
        List of subscriptions

    """
    _range: Union[List[int], None] = list(map(int, range.split(","))) if range else None
    _sort: Union[List[str], None] = sort.split(",") if sort else None
    _filter: Union[List[str], None] = filter.split(",") if filter else None
    logger.info("subscriptions_filterable() called", range=_range, sort=_sort, filter=_filter)
    query = SubscriptionTable.query.join(SubscriptionTable.product).options(
        contains_eager(SubscriptionTable.product), defer("product_id")
    )
    query_result = _query_with_filters(response, query, _range, _sort, _filter)
    return query_result


@router.get(
    "/workflows/{subscription_id}", response_model=SubscriptionWorkflowListsSchema, response_model_exclude_none=True
)
def subscription_workflows_by_id(subscription_id: UUID) -> Dict[str, List[Dict[str, Union[List[Any], str]]]]:
    subscription = SubscriptionTable.query.options(joinedload("product"), joinedload("product.workflows")).get(
        subscription_id
    )
    if not subscription:
        raise_status(HTTPStatus.NOT_FOUND)

    return subscription_workflows(subscription)



import difflib
import pprint
from copy import deepcopy
from functools import reduce
from itertools import chain, repeat
from typing import Callable, Dict, List, Optional, Tuple, Union, cast
from uuid import uuid4

import structlog

from orchestrator.config import IMS_CIRCUIT_ID, PORT_SUBSCRIPTION_ID
from orchestrator.db import ProcessTable
from orchestrator.forms import post_process
from orchestrator.services.processes import StateMerger, _db_create_process
from orchestrator.types import FormGenerator, InputForm, State
from orchestrator.utils.errors import error_state_to_dict
from orchestrator.utils.json import json_dumps, json_loads
from orchestrator.workflow import Process as WFProcess
from orchestrator.workflow import ProcessStat, Step, Success, Workflow, runwf
from orchestrator.workflows import ALL_WORKFLOWS, LazyWorkflowInstance, get_workflow

logger = structlog.get_logger(__name__)


def _raise_exception(state):
    if isinstance(state, Exception):
        raise state
    return state


def assert_success(result):
    assert (
        result.on_failed(_raise_exception).on_waiting(_raise_exception).issuccess()
    ), f"Unexpected process status. Expected Success, but was: {result}"


def assert_waiting(result):
    assert result.on_failed(
        _raise_exception
    ).iswaiting(), f"Unexpected process status. Expected Waiting, but was: {result}"


def assert_suspended(result):
    assert result.on_failed(
        _raise_exception
    ).issuspend(), f"Unexpected process status. Expected Suspend, but was: {result}"


def assert_aborted(result):
    assert result.on_failed(_raise_exception).isabort(), f"Unexpected process status. Expected Abort, but was: {result}"


def assert_failed(result):
    assert result.isfailed(), f"Unexpected process status. Expected Failed, but was: {result}"


def assert_complete(result):
    assert result.on_failed(
        _raise_exception
    ).iscomplete(), f"Unexpected process status. Expected Complete, but was: {result}"


def assert_state(result, expected):
    state = result.unwrap()
    actual = {}
    for key in expected.keys():
        actual[key] = state[key]
    assert expected == actual, f"Invalid state. Expected superset of: {expected}, but was: {actual}"


def assert_state_equal(result: ProcessTable, expected: Dict, excluded_keys: Optional[List[str]] = None) -> None:
    """Test state with certain keys excluded from both actual and expected state."""
    if excluded_keys is None:
        excluded_keys = ["process_id", "workflow_target", "workflow_name"]
    state = deepcopy(extract_state(result))
    expected_state = deepcopy(expected)
    for key in excluded_keys:
        if key in state:
            del state[key]
        if key in expected_state:
            del expected_state[key]

    assert state == expected_state, "Unexpected state:\n" + "\n".join(
        difflib.ndiff(pprint.pformat(state).splitlines(), pprint.pformat(expected_state).splitlines())
    )


def assert_assignee(log, expected):
    actual = log[-1][0].assignee
    assert expected == actual, f"Unexpected assignee. Expected {expected}, but was: {actual}"


def assert_step_name(log, expected):
    actual = log[-1][0]
    assert actual.name == expected, f"Unexpected name. Expected {expected}, but was: {actual}"


def extract_state(result):
    return result.on_failed(error_state_to_dict).on_waiting(error_state_to_dict).unwrap()


def extract_error(result):
    from orchestrator.workflow import Process

    assert isinstance(result, Process), f"Expected a Process, but got {repr(result)} of type {type(result)}"
    assert not isinstance(result.s, Process), "Result contained a Process in a Process, this should not happen"

    return extract_state(result).get("error")


ALL_TEST_WORKFLOWS = set()


class WorkflowInstanceForTests(LazyWorkflowInstance):
    """Register Test workflows.

    Similar to `LazyWorkflowInstance` but does not require an import during instantiate
    Used for creating test workflows
    """

    package: str
    function: str
    is_callable: bool

    def __init__(self, workflow: Workflow, name: str) -> None:
        self.workflow = workflow
        self.name = name
        ALL_WORKFLOWS[name] = self
        ALL_TEST_WORKFLOWS.add(name)

    def instantiate(self) -> Workflow:
        """Import and instantiate a workflow and return it.

        This can be as simple as merely importing a workflow function. However, if it concerns a workflow generating
        function, that function will be called with or without arguments as specified.

        Returns:
            A workflow function.

        """
        self.workflow.name = self.name
        return self.workflow

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"WorkflowInstanceForTests('{self.workflow}','{self.name}')"


def _store_step(step_log: List[Tuple[Step, WFProcess]]) -> Callable[[ProcessStat, Step, WFProcess], WFProcess]:
    def __store_step(pstat: ProcessStat, step: Step, state: WFProcess) -> WFProcess:
        try:
            state = state.map(lambda s: json_loads(json_dumps(s)))
        except Exception:
            logger.exception("Step state is not valid json", state=state)
        step_log.append((step, state))
        return state

    return __store_step


def _sanitize_input(input_data: Union[State, List[State]]) -> List[State]:
    # To be backwards compatible convert single dict to list
    if not isinstance(input_data, List):
        input_data = [input_data]

    # We need a copy here and we want to mimic the actual code that returns a serialized version of the state
    return cast(List[State], json_loads(json_dumps(input_data)))


def run_workflow(workflow_key: str, input_data: Union[State, List[State]]) -> Tuple[WFProcess, ProcessStat, List]:
    # ATTENTION!! This code needs to be as similar as possible to `server.services.processes.start_process`
    # The main differences are: we use a different step log function and we don't run in
    # a sepperate thread
    user_data = _sanitize_input(input_data)
    user = "john.doe"

    step_log: List[Tuple[Step, WFProcess]] = []

    pid = uuid4()
    workflow = get_workflow(workflow_key)
    initial_state = {
        "process_id": pid,
        "reporter": user,
        "workflow_name": workflow_key,
        "workflow_target": workflow.target,
    }

    user_input = post_process(workflow.initial_input_form, initial_state, user_data)

    pstat = ProcessStat(
        pid, workflow=workflow, state=Success({**user_input, **initial_state}), log=workflow.steps, current_user=user
    )

    _db_create_process(pstat)

    result = runwf(pstat, _store_step(step_log))

    return result, pstat, step_log


def resume_workflow(
    process: ProcessStat, step_log: List[Tuple[Step, WFProcess]], input_data: State
) -> Tuple[WFProcess, List]:
    # ATTENTION!! This code needs to be as similar as possible to `server.services.processes.resume_process`
    # The main differences are: we use a different step log function and we don't run in
    # a sepperate thread
    user_data = _sanitize_input(input_data)

    persistent = list(filter(lambda p: not (p[1].isfailed() or p[1].issuspend() or p[1].iswaiting()), step_log))
    nr_of_steps_done = len(persistent)
    remaining_steps = process.workflow.steps[nr_of_steps_done:]

    _, current_state = step_log[-1]

    user_input = post_process(remaining_steps[0].form, current_state.unwrap(), user_data)
    state = current_state.map(lambda state: StateMerger.merge(deepcopy(state), user_input))

    updated_process = process.update(log=remaining_steps, state=state)
    result = runwf(updated_process, _store_step(step_log))
    return result, step_log


def assert_product_blocks_equal(expected, actual):
    def key_value_sort(item):
        name = list(item.keys())[0]
        if name == "Virtual Circuit":
            items = list(filter(lambda k: list(k.keys())[0] == IMS_CIRCUIT_ID, item["Virtual Circuit"]))
            id_ = items[0][IMS_CIRCUIT_ID]
            return f"1_{id_}"
        if name == "Service Attach Point":
            port_subscription_id = list(
                filter(lambda k: list(k.keys())[0] == PORT_SUBSCRIPTION_ID, item["Service Attach Point"])
            )
            port_subscription_id = port_subscription_id[0][PORT_SUBSCRIPTION_ID]
            vlanrange = list(filter(lambda k: list(k.keys())[0] == "vlanrange", item["Service Attach Point"]))
            vlanrange = vlanrange[0]["vlanrange"]
            return f"2_{port_subscription_id}_{vlanrange}"
        return name

    expected.sort(key=key_value_sort)
    actual.sort(key=key_value_sort)

    def accumulate_list_of_tuples(acc: List, dikt: Dict) -> List[Tuple]:
        acc.extend(dikt.items())
        return acc

    expected_product_block_names = [list(p.keys())[0] for p in expected]
    actual_product_block_names = [list(p.keys())[0] for p in actual]
    assert (
        expected_product_block_names == actual_product_block_names
    ), f"Expected the following product blocks: {expected_product_block_names}, but got {actual_product_block_names}"

    for expected_pb, actual_pb in zip(expected, actual):
        for expected_instance_values, actual_instance_values in zip(expected_pb.values(), actual_pb.values()):
            expected_tuples = set(reduce(accumulate_list_of_tuples, expected_instance_values, []))
            actual_tuples = set(reduce(accumulate_list_of_tuples, actual_instance_values, []))
            missing_instance_values = expected_tuples - actual_tuples
            unexpected_instance_values = actual_tuples - expected_tuples
            assert (
                not missing_instance_values
            ), f"Missing instance value(s): {missing_instance_values}; Unexpected: {unexpected_instance_values}"
            assert not unexpected_instance_values, f"Unexpected instance values: {unexpected_instance_values}"


def run_form_generator(
    form_generator: FormGenerator, extra_inputs: Optional[List[State]] = None
) -> Tuple[List[dict], State]:
    """Run a form generator to get the resulting forms and result.

    Warning! This does not run the actual pydantic validation on purpose. However you should
    make sure that anything in extra_inputs matched the values and types as if the pydantic validation has
    been ran.

    Args:
        form_generator: A form generator
        extra_inputs: Optional list of user input dicts for each page in the generator.
                      If no input is given for a page an empty dict is used.
                      The default value from the form is used as default value for a field.
    Returns:
        A list of generated forms and the result state for the whole generator.

    Example:
        Given the following form generator:

        >>> from orchestrator.forms import FormPage
        >>> def form_generator(state):
        ...     class TestForm(FormPage):
        ...         field: str = "foo"
        ...     user_input = yield TestForm
        ...     return {**user_input.dict(), "bar": 42}

        You can run this without extra_inputs
        >>> forms, result = run_form_generator(form_generator({"state_field": 1}))
        >>> forms
        [{'title': 'unknown', 'type': 'object', 'properties': {'field': {'title': 'Field', 'default': 'foo', 'type': 'string'}}, 'additionalProperties': False}]
        >>> result
        {'field': 'foo', 'bar': 42}


        Or with extra_inputs:
        >>> forms, result = run_form_generator(form_generator({'state_field': 1}), [{'field':'baz'}])
        >>> forms
        [{'title': 'unknown', 'type': 'object', 'properties': {'field': {'title': 'Field', 'default': 'foo', 'type': 'string'}}, 'additionalProperties': False}]
        >>> result
        {'field': 'baz', 'bar': 42}

    """
    forms: List[dict] = []
    result: State = {"s": 3}
    if extra_inputs is None:
        extra_inputs = []

    try:
        form = cast(InputForm, next(form_generator))
        forms.append(form.schema())
        for extra_input in chain(extra_inputs, repeat(cast(State, {}))):
            user_input_data = {field_name: field.default for field_name, field in form.__fields__.items()}
            user_input_data.update(extra_input)
            user_input = form.construct(**user_input_data)
            form = form_generator.send(user_input)
            forms.append(form.schema())
    except StopIteration as stop:
        result = stop.value

    return forms, result
